"""Tests for web capabilities module"""

import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from cody.core.web import (
    html_to_markdown,
    webfetch,
    websearch,
    _parse_ddg_results,
)


# ── html_to_markdown ─────────────────────────────────────────────────────────


def test_html_to_markdown_headings():
    html = "<h1>Title</h1><h2>Subtitle</h2><h3>Section</h3>"
    md = html_to_markdown(html)
    assert "# Title" in md
    assert "## Subtitle" in md
    assert "### Section" in md


def test_html_to_markdown_paragraphs():
    html = "<p>First paragraph</p><p>Second paragraph</p>"
    md = html_to_markdown(html)
    assert "First paragraph" in md
    assert "Second paragraph" in md


def test_html_to_markdown_links():
    html = '<a href="https://example.com">Click here</a>'
    md = html_to_markdown(html)
    assert "[Click here](https://example.com)" in md


def test_html_to_markdown_bold_italic():
    html = "<strong>bold</strong> and <em>italic</em>"
    md = html_to_markdown(html)
    assert "**bold**" in md
    assert "*italic*" in md


def test_html_to_markdown_code():
    html = "<code>inline code</code>"
    md = html_to_markdown(html)
    assert "`inline code`" in md


def test_html_to_markdown_pre():
    html = "<pre>code block</pre>"
    md = html_to_markdown(html)
    assert "```" in md
    assert "code block" in md


def test_html_to_markdown_list():
    html = "<ul><li>Item 1</li><li>Item 2</li></ul>"
    md = html_to_markdown(html)
    assert "- Item 1" in md
    assert "- Item 2" in md


def test_html_to_markdown_skips_script():
    html = "<p>Visible</p><script>alert('hidden')</script><p>Also visible</p>"
    md = html_to_markdown(html)
    assert "Visible" in md
    assert "Also visible" in md
    assert "alert" not in md


def test_html_to_markdown_skips_style():
    html = "<style>body { color: red; }</style><p>Content</p>"
    md = html_to_markdown(html)
    assert "Content" in md
    assert "color" not in md


def test_html_to_markdown_blockquote():
    html = "<blockquote>Quoted text</blockquote>"
    md = html_to_markdown(html)
    assert "> Quoted text" in md


def test_html_to_markdown_empty():
    assert html_to_markdown("") == ""


def test_html_to_markdown_br():
    html = "Line1<br>Line2"
    md = html_to_markdown(html)
    assert "Line1\nLine2" in md


# ── webfetch ─────────────────────────────────────────────────────────────────


@pytest.mark.asyncio
async def test_webfetch_html():
    mock_resp = MagicMock()
    mock_resp.text = "<html><body><h1>Hello</h1><p>World</p></body></html>"
    mock_resp.headers = {"content-type": "text/html; charset=utf-8"}
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.get = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await webfetch("https://example.com")

    assert "Hello" in result
    assert "World" in result


@pytest.mark.asyncio
async def test_webfetch_json():
    mock_resp = MagicMock()
    mock_resp.text = '{"key": "value"}'
    mock_resp.json = MagicMock(return_value={"key": "value"})
    mock_resp.headers = {"content-type": "application/json"}
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.get = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await webfetch("https://api.example.com/data")

    assert '"key"' in result
    assert '"value"' in result


@pytest.mark.asyncio
async def test_webfetch_plain_text():
    mock_resp = MagicMock()
    mock_resp.text = "Just plain text"
    mock_resp.headers = {"content-type": "text/plain"}
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.get = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await webfetch("https://example.com/file.txt")

    assert result == "Just plain text"


@pytest.mark.asyncio
async def test_webfetch_unsupported_content():
    mock_resp = MagicMock()
    mock_resp.text = "binary stuff"
    mock_resp.headers = {"content-type": "application/octet-stream"}
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.get = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await webfetch("https://example.com/file.bin")

    assert "Unsupported content type" in result


# ── websearch ────────────────────────────────────────────────────────────────


@pytest.mark.asyncio
async def test_websearch():
    mock_html = """
    <div class="result">
        <a class="result__a" href="https://example.com">Example Title</a>
        <a class="result__snippet">Example description of the result</a>
    </div>
    """
    mock_resp = MagicMock()
    mock_resp.text = mock_html
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.post = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await websearch("test query")

    assert "Search results for: test query" in result
    assert "Example Title" in result


@pytest.mark.asyncio
async def test_websearch_no_results():
    mock_resp = MagicMock()
    mock_resp.text = "<html><body>No results</body></html>"
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.post = AsyncMock(return_value=mock_resp)
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)

    with patch("cody.core.web.httpx.AsyncClient", return_value=mock_client):
        result = await websearch("xyznonexistent123")

    assert "No search results found" in result


# ── _parse_ddg_results ───────────────────────────────────────────────────────


def test_parse_ddg_results_basic():
    html = """
    <a class="result__a" href="https://example.com">Title One</a>
    <a class="result__snippet">Snippet one text</a>
    <a class="result__a" href="https://example2.com">Title Two</a>
    <a class="result__snippet">Snippet two text</a>
    """
    results = _parse_ddg_results(html, max_results=10)
    assert len(results) == 2
    assert results[0]["title"] == "Title One"
    assert results[0]["url"] == "https://example.com"
    assert results[0]["snippet"] == "Snippet one text"


def test_parse_ddg_results_max_limit():
    html = """
    <a class="result__a" href="https://a.com">A</a>
    <a class="result__snippet">Snippet A</a>
    <a class="result__a" href="https://b.com">B</a>
    <a class="result__snippet">Snippet B</a>
    <a class="result__a" href="https://c.com">C</a>
    <a class="result__snippet">Snippet C</a>
    """
    results = _parse_ddg_results(html, max_results=2)
    assert len(results) == 2


def test_parse_ddg_results_empty():
    results = _parse_ddg_results("<html></html>", max_results=10)
    assert results == []


def test_parse_ddg_results_uddg_redirect():
    html = """
    <a class="result__a" href="/l/?uddg=https%3A%2F%2Freal.com%2Fpage&rut=abc">Title</a>
    <a class="result__snippet">Snippet</a>
    """
    results = _parse_ddg_results(html, max_results=10)
    assert len(results) == 1
    assert results[0]["url"] == "https://real.com/page"



"""Tool endpoint — POST /tool.

Migrated from cody/server.py. Direct tool invocation without Agent.
"""

from pathlib import Path

from fastapi import APIRouter

from cody.core.deps import ToolContext
from cody.core.errors import (
    CodyAPIError, ErrorCode,
    ToolError, ToolPermissionDenied, ToolPathDenied, ToolInvalidParams,
)

from ..helpers import raise_structured
from ..models import ToolRequest, ToolResponse
from ..state import get_config, create_full_deps

router = APIRouter(tags=["tool"])


@router.post("/tool", response_model=ToolResponse)
async def call_tool(request: ToolRequest):
    """Call a tool directly."""
    from cody.core import tools

    tool_func = getattr(tools, request.tool, None)
    if not tool_func:
        raise_structured(
            ErrorCode.TOOL_NOT_FOUND,
            f"Tool not found: {request.tool}",
            status_code=404,
        )

    try:
        workdir = Path(request.workdir) if request.workdir else Path.cwd()
        config = get_config(workdir)
        deps = create_full_deps(config, workdir)

        ctx = ToolContext(deps)
        result = await tool_func(ctx, **request.params)

        return ToolResponse(result=result)

    except (ToolPermissionDenied, ToolPathDenied) as e:
        raise_structured(e.code, e.message, status_code=403)
    except ToolInvalidParams as e:
        raise_structured(e.code, e.message, status_code=400)
    except ToolError as e:
        raise_structured(e.code, e.message, status_code=500)
    except CodyAPIError:
        raise
    except Exception as e:
        raise_structured(
            ErrorCode.TOOL_ERROR, str(e), status_code=500,
            details={"tool": request.tool},
        )



"""Search tools — grep, glob, patch, search_files."""

import fnmatch
import re
from pathlib import Path

from pydantic_ai import RunContext

from ..deps import CodyDeps
from ..errors import ToolInvalidParams
from ._base import _audit_tool_call, _check_permission, _resolve_and_check
from ._file_filter import (
    _DEFAULT_IGNORE_FILES,
    _is_binary,
    _is_gitignored,
    _is_ignored_dir,
    _iter_files,
    _parse_gitignore,
)

_MAX_FILE_SIZE = 1_048_576  # 1 MB — skip files larger than this in Python grep


async def grep(
    ctx: RunContext['CodyDeps'],
    pattern: str,
    path: str = ".",
    include: str = "",
) -> str:
    """Search file contents using a regular expression

    Args:
        pattern: Regular expression pattern to search for
        path: Directory or file to search in (relative or absolute)
        include: Optional glob to filter filenames (e.g. "*.py")
    """
    full_path = _resolve_and_check(
        ctx.deps.workdir, path, allow_read_outside=not ctx.deps.strict_read_boundary, allowed_roots=ctx.deps.allowed_roots
    )

    if not full_path.exists():
        raise ToolInvalidParams(f"Path not found: {path}")

    try:
        regex = re.compile(pattern)
    except re.error as e:
        raise ToolInvalidParams(f"Invalid regex pattern: {e}")

    max_matches = 200
    workdir_resolved = ctx.deps.workdir.resolve()

    # Python implementation
    matches: list[str] = []
    skipped_large: list[str] = []

    if full_path.is_file():
        files = [full_path]
    else:
        gitignore_patterns = _parse_gitignore(workdir_resolved)
        files = _iter_files(full_path, workdir_resolved, gitignore_patterns)

    for file_path in files:
        if not file_path.is_file():
            continue
        if include and not fnmatch.fnmatch(file_path.name, include):
            continue
        if _is_binary(file_path):
            continue

        # Skip files larger than _MAX_FILE_SIZE but record them
        try:
            file_size = file_path.stat().st_size
            if file_size > _MAX_FILE_SIZE:
                rel = file_path.relative_to(workdir_resolved)
                size_mb = file_size / 1_048_576
                skipped_large.append(f"{rel} ({size_mb:.1f} MB)")
                continue
        except OSError:
            continue

        try:
            content = file_path.read_text(errors="ignore")
        except (PermissionError, OSError):
            continue

        rel = file_path.relative_to(workdir_resolved)
        for line_no, line in enumerate(content.splitlines(), 1):
            if regex.search(line):
                matches.append(f"{rel}:{line_no}: {line.rstrip()}")
                if len(matches) >= max_matches:
                    matches.append(f"... (truncated at {max_matches} matches)")
                    if skipped_large:
                        matches.append(
                            f"\n[NOTE] {len(skipped_large)} large file(s) skipped "
                            f"(>{_MAX_FILE_SIZE // 1_048_576} MB): "
                            + ", ".join(skipped_large[:10])
                        )
                    return "\n".join(matches)

    if not matches and not skipped_large:
        return f"No matches found for pattern: {pattern}"

    lines = matches or [f"No matches found for pattern: {pattern}"]
    if skipped_large:
        lines.append(
            f"\n[NOTE] {len(skipped_large)} large file(s) skipped "
            f"(>{_MAX_FILE_SIZE // 1_048_576} MB): "
            + ", ".join(skipped_large[:10])
        )
        if len(skipped_large) > 10:
            lines.append(f"  ... and {len(skipped_large) - 10} more")

    return "\n".join(lines)


async def glob(
    ctx: RunContext['CodyDeps'],
    pattern: str,
    path: str = ".",
) -> str:
    """Find files by glob pattern

    Args:
        pattern: Relative glob pattern (e.g. "**/*.py", "*.txt", "src/**/*.ts").
            Must be a relative pattern — absolute paths are NOT allowed.
            Use the 'path' parameter to specify the search directory instead.
        path: Base directory to search from (relative or absolute)
    """
    if Path(pattern).is_absolute():
        raise ToolInvalidParams(
            f"Glob pattern must be relative, not absolute: {pattern}. "
            f"Use the 'path' parameter to specify the search directory."
        )

    full_path = _resolve_and_check(
        ctx.deps.workdir, path, allow_read_outside=not ctx.deps.strict_read_boundary, allowed_roots=ctx.deps.allowed_roots
    )

    if not full_path.exists():
        raise ToolInvalidParams(f"Path not found: {path}")

    if not full_path.is_dir():
        raise ToolInvalidParams(f"Not a directory: {path}")

    workdir_resolved = ctx.deps.workdir.resolve()
    gitignore_patterns = _parse_gitignore(workdir_resolved)
    results: list[str] = []
    max_results = 500
    all_roots = [workdir_resolved] + [r.resolve() for r in (ctx.deps.allowed_roots or [])]

    for match in sorted(full_path.glob(pattern)):
        # Security: skip anything outside all permitted roots
        resolved_match = match.resolve()
        if not any(resolved_match.is_relative_to(root) for root in all_roots):
            continue

        try:
            rel = match.relative_to(workdir_resolved)
        except ValueError:
            # File is in an allowed_root outside workdir — show absolute path
            rel = match
        rel_posix = str(rel).replace('\\', '/')
        is_dir = match.is_dir()

        # Check if any parent directory is ignored
        skip = False
        for part in rel.parts[:-1]:
            if _is_ignored_dir(part):
                skip = True
                break
        if skip:
            continue

        # Check the item itself
        if is_dir and _is_ignored_dir(match.name):
            continue
        if not is_dir and match.name in _DEFAULT_IGNORE_FILES:
            continue

        # Check gitignore
        if gitignore_patterns and _is_gitignored(rel_posix, gitignore_patterns, is_dir=is_dir):
            continue

        prefix = "[dir]" if is_dir else "[file]"
        results.append(f"{prefix} {rel}")
        if len(results) >= max_results:
            results.append(f"... (truncated at {max_results} results)")
            break

    if not results:
        return f"No files matched pattern: {pattern}"

    return "\n".join(results)


async def patch(
    ctx: RunContext['CodyDeps'],
    path: str,
    diff: str,
) -> str:
    """Apply a unified diff patch to a file

    Args:
        path: Path to the file to patch
        diff: Unified diff content (lines starting with +/- and context lines)
    """
    await _check_permission(ctx, "patch", args_summary=path)
    full_path = _resolve_and_check(ctx.deps.workdir, path, allowed_roots=ctx.deps.allowed_roots)

    if not full_path.exists():
        raise ToolInvalidParams(f"File not found: {path}")

    original_lines = full_path.read_text(encoding="utf-8", errors="replace").splitlines(keepends=True)
    result_lines: list[str] = []
    orig_idx = 0

    diff_lines = diff.splitlines(keepends=True)
    i = 0

    while i < len(diff_lines):
        line = diff_lines[i]

        # Skip diff headers
        if line.startswith("---") or line.startswith("+++"):
            i += 1
            continue

        # Parse hunk header
        if line.startswith("@@"):
            hunk_match = re.match(r"@@ -(\d+)(?:,\d+)? \+\d+(?:,\d+)? @@", line)
            if not hunk_match:
                raise ToolInvalidParams(f"Invalid hunk header: {line.rstrip()}")
            hunk_start = int(hunk_match.group(1)) - 1  # 0-indexed
            if hunk_start < 0 or hunk_start > len(original_lines):
                raise ToolInvalidParams(
                    f"Invalid hunk start line {hunk_start + 1}: "
                    f"file has {len(original_lines)} lines"
                )

            # Copy lines before this hunk
            while orig_idx < hunk_start:
                if orig_idx < len(original_lines):
                    result_lines.append(original_lines[orig_idx])
                orig_idx += 1

            i += 1
            # Process hunk lines
            while i < len(diff_lines) and not diff_lines[i].startswith("@@"):
                dl = diff_lines[i]
                if dl.startswith("\\"):
                    # "\ No newline at end of file" marker
                    if result_lines and result_lines[-1].endswith("\n"):
                        result_lines[-1] = result_lines[-1][:-1]
                    i += 1
                    continue
                elif dl.startswith("-"):
                    # Remove line — skip it from original
                    orig_idx += 1
                elif dl.startswith("+"):
                    # Add line
                    content = dl[1:]
                    if not content.endswith("\n"):
                        content += "\n"
                    result_lines.append(content)
                elif dl.startswith(" ") or dl.strip() == "":
                    # Context line — validate it matches original
                    if orig_idx < len(original_lines):
                        expected = original_lines[orig_idx]
                        context_content = dl[1:] if dl.startswith(" ") else dl
                        if context_content.rstrip("\n") != expected.rstrip("\n"):
                            raise ToolInvalidParams(
                                f"Context mismatch at line {orig_idx + 1}: "
                                f"expected {expected.rstrip()!r}, "
                                f"got {context_content.rstrip()!r}"
                            )
                        result_lines.append(original_lines[orig_idx])
                    orig_idx += 1
                else:
                    break
                i += 1
        else:
            i += 1

    # Copy remaining lines
    while orig_idx < len(original_lines):
        result_lines.append(original_lines[orig_idx])
        orig_idx += 1

    patched_content = "".join(result_lines)
    full_path.write_text(patched_content, encoding="utf-8")

    # Track in file history
    if ctx.deps.file_history:
        original_content = "".join(original_lines)
        ctx.deps.file_history.record(path, original_content, patched_content, operation="patch")

    _audit_tool_call(ctx, "file_edit", "patch", f"path={path}", f"Patched {path}")

    return f"Patched {path} successfully"


async def search_files(
    ctx: RunContext['CodyDeps'],
    query: str,
    path: str = ".",
) -> str:
    """Search for files by name (fuzzy match)

    Args:
        query: Search query to match against file names
        path: Directory to search in (relative or absolute)
    """
    full_path = _resolve_and_check(
        ctx.deps.workdir, path, allow_read_outside=not ctx.deps.strict_read_boundary, allowed_roots=ctx.deps.allowed_roots
    )

    if not full_path.exists():
        raise ToolInvalidParams(f"Path not found: {path}")

    if not full_path.is_dir():
        raise ToolInvalidParams(f"Not a directory: {path}")

    workdir_resolved = ctx.deps.workdir.resolve()
    gitignore_patterns = _parse_gitignore(workdir_resolved)
    query_lower = query.lower()

    results: list[tuple[int, str]] = []
    max_results = 100

    for file_path in _iter_files(full_path, workdir_resolved, gitignore_patterns):
        name_lower = file_path.name.lower()
        rel = str(file_path.relative_to(workdir_resolved))

        # Scoring: exact match > starts with > contains > path contains
        if name_lower == query_lower:
            score = 0
        elif name_lower.startswith(query_lower):
            score = 1
        elif query_lower in name_lower:
            score = 2
        elif query_lower in rel.lower():
            score = 3
        else:
            continue

        results.append((score, rel))

    results.sort()
    output = [r[1] for r in results[:max_results]]

    if not output:
        return f"No files found matching: {query}"

    if len(results) > max_results:
        output.append(f"... ({len(results) - max_results} more)")

    return "\n".join(output)



"""CLI session management commands."""

from pathlib import Path

import click
from rich.markdown import Markdown
from rich.panel import Panel

from ...sdk.client import CodyClient
from ..utils import console


def _make_client() -> CodyClient:
    return CodyClient(workdir=str(Path.cwd()))


@click.group()
def sessions():
    """Manage chat sessions"""


@sessions.command('list')
@click.option('--limit', default=20, help='Number of sessions to show')
def sessions_list(limit):
    """List recent chat sessions"""
    client = _make_client()
    all_sessions = client.list_sessions(limit=limit)

    if not all_sessions:
        console.print("[yellow]No sessions found[/yellow]")
        return

    console.print("[bold]Recent sessions:[/bold]\n")
    for s in all_sessions:
        console.print(
            f"  [bold]{s.id}[/bold]  {s.title[:50]:<50}  "
            f"[dim]{s.message_count} msgs  {s.updated_at[:10]}[/dim]"
        )


@sessions.command('show')
@click.argument('session_id')
def sessions_show(session_id):
    """Show a session's conversation"""
    client = _make_client()
    try:
        session = client.get_session(session_id)
    except Exception:
        console.print(f"[red]Session not found: {session_id}[/red]")
        return

    console.print(
        Panel(
            f"Title: {session.title}\n"
            f"Model: {session.model}\n"
            f"Workdir: {session.workdir}\n"
            f"Created: {session.created_at[:19]}\n"
            f"Messages: {session.message_count}",
            title=f"[bold]Session {session.id}[/bold]",
            border_style="blue",
        )
    )

    for msg in session.messages:
        if msg["role"] == "user":
            console.print(f"\n[bold blue]You:[/bold blue] {msg['content']}")
        else:
            console.print("\n[bold green]Cody:[/bold green]")
            console.print(Markdown(msg["content"]))


@sessions.command('delete')
@click.argument('session_id')
@click.confirmation_option(prompt='Are you sure you want to delete this session?')
def sessions_delete(session_id):
    """Delete a chat session"""
    client = _make_client()
    try:
        client.delete_session(session_id)
        console.print(f"[green]Deleted session: {session_id}[/green]")
    except Exception:
        console.print(f"[red]Session not found: {session_id}[/red]")



"""CLI skills management commands."""

from pathlib import Path

import click
from rich.markdown import Markdown
from rich.panel import Panel

from ...sdk.client import CodyClient
from ..utils import console


def _make_client() -> CodyClient:
    return CodyClient(workdir=str(Path.cwd()))


@click.group()
def skills():
    """Manage skills"""


@skills.command('list')
def skills_list():
    """List all available skills"""
    client = _make_client()
    all_skills = client.list_skills()

    if not all_skills:
        console.print("[yellow]No skills found[/yellow]")
        return

    console.print("[bold]Available Skills:[/bold]\n")

    for skill in sorted(all_skills, key=lambda s: s["name"]):
        status = "[green]on[/green]" if skill["enabled"] else "[red]off[/red]"
        source = f"[dim]({skill['source']})[/dim]"
        console.print(f"  [{status}] {skill['name']} {source}")
        console.print(f"        {skill['description']}")


@skills.command('show')
@click.argument('skill_name')
def skills_show(skill_name):
    """Show skill documentation"""
    client = _make_client()
    try:
        skill = client.get_skill(skill_name)
    except Exception:
        console.print(f"[red]Skill not found: {skill_name}[/red]")
        return

    md = Markdown(skill["documentation"])
    console.print(Panel(md, title=f"[bold]{skill['name']}[/bold]", border_style="blue"))


@skills.command('enable')
@click.argument('skill_name')
def skills_enable(skill_name):
    """Enable a skill"""
    client = _make_client()
    try:
        client.enable_skill(skill_name)
        console.print(f"[green]Enabled skill: {skill_name}[/green]")
    except Exception:
        console.print(f"[red]Skill not found: {skill_name}[/red]")


@skills.command('disable')
@click.argument('skill_name')
def skills_disable(skill_name):
    """Disable a skill"""
    client = _make_client()
    try:
        client.disable_skill(skill_name)
        console.print(f"[green]Disabled skill: {skill_name}[/green]")
    except Exception:
        console.print(f"[red]Skill not found: {skill_name}[/red]")



"""Tests for core/prompts.py — system prompt construction."""

from cody.core.prompts import build_system_prompt


class TestBuildSystemPrompt:
    """Verify that build_system_prompt() produces a complete prompt."""

    def test_returns_string(self):
        result = build_system_prompt()
        assert isinstance(result, str)
        assert len(result) > 0

    def test_contains_role_definition(self):
        result = build_system_prompt()
        assert "You are Cody" in result

    def test_contains_capabilities(self):
        result = build_system_prompt()
        assert "## Capabilities" in result
        assert "file operations" in result
        assert "shell command execution" in result
        assert "sub-agent spawning" in result

    def test_contains_boundaries(self):
        result = build_system_prompt()
        assert "## Boundaries" in result
        assert "NEVER delete files outside the project directory" in result
        assert "NEVER modify system files" in result
        assert "NEVER run destructive commands" in result

    def test_contains_output_format(self):
        result = build_system_prompt()
        assert "## Output Format" in result
        assert "Markdown" in result

    def test_contains_code_quality(self):
        result = build_system_prompt()
        assert "## Code Quality" in result
        assert "Match existing code style" in result

    def test_contains_task_completion(self):
        result = build_system_prompt()
        assert "## Task Completion" in result
        assert "tests pass" in result

    def test_contains_context_management(self):
        result = build_system_prompt()
        assert "## Context Management" in result
        assert "save_memory()" in result

    def test_contains_thinking_guidance(self):
        result = build_system_prompt()
        assert "## Approach" in result
        assert "Understand" in result
        assert "Plan" in result
        assert "Execute" in result
        assert "Verify" in result
        assert "Report" in result
        assert "Do NOT skip step 1" in result

    def test_contains_sub_agent_guidance(self):
        result = build_system_prompt()
        assert "## Sub-Agent Parallelism" in result
        assert "spawn_agent()" in result

    def test_contains_skills_guidance(self):
        result = build_system_prompt()
        assert "## Skills Usage" in result
        assert "read_skill(skill_name)" in result
        assert "context overhead" in result

    def test_sections_order(self):
        """Verify sections appear in the expected order."""
        result = build_system_prompt()
        capabilities_pos = result.index("## Capabilities")
        approach_pos = result.index("## Approach")
        sub_agent_pos = result.index("## Sub-Agent Parallelism")
        skills_pos = result.index("## Skills Usage")

        assert capabilities_pos < approach_pos < sub_agent_pos < skills_pos



"""Tests for cody/core/project_instructions.py"""



from cody.core.project_instructions import (
    CODY_MD_FILENAME,
    CODY_MD_TEMPLATE,
    load_project_instructions,
)


# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------


def test_cody_md_filename():
    assert CODY_MD_FILENAME == "CODY.md"


def test_cody_md_template_is_non_empty():
    assert CODY_MD_TEMPLATE.strip()


def test_cody_md_template_is_string():
    assert isinstance(CODY_MD_TEMPLATE, str)


# ---------------------------------------------------------------------------
# load_project_instructions — no files present
# ---------------------------------------------------------------------------


def test_returns_none_when_no_files(tmp_path, monkeypatch):
    """Returns None when neither global nor project CODY.md exists."""
    monkeypatch.setenv("HOME", str(tmp_path))
    result = load_project_instructions(tmp_path / "project")
    assert result is None


# ---------------------------------------------------------------------------
# load_project_instructions — project-only
# ---------------------------------------------------------------------------


def test_returns_project_content_only(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("# My Project\nHello.", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result == "# My Project\nHello."


def test_strips_leading_trailing_whitespace_project(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("\n\n# Project\n\n", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result == "# Project"


def test_empty_project_file_returns_none(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("   \n\n   ", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result is None


# ---------------------------------------------------------------------------
# load_project_instructions — global-only
# ---------------------------------------------------------------------------


def test_returns_global_content_only(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("# Global Instructions", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()

    result = load_project_instructions(project_dir)
    assert result == "# Global Instructions"


def test_empty_global_file_returns_none(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()

    result = load_project_instructions(project_dir)
    assert result is None


# ---------------------------------------------------------------------------
# load_project_instructions — both files present
# ---------------------------------------------------------------------------


def test_merges_global_and_project(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("# Global", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("# Project", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result == "# Global\n\n---\n\n# Project"


def test_global_comes_before_project(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("GLOBAL_CONTENT", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("PROJECT_CONTENT", encoding="utf-8")

    result = load_project_instructions(project_dir)
    global_idx = result.index("GLOBAL_CONTENT")
    project_idx = result.index("PROJECT_CONTENT")
    assert global_idx < project_idx


def test_separator_present_when_both_files(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("A", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("B", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert "---" in result


def test_no_separator_when_only_one_file(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("Only project", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert "---" not in result


def test_empty_global_does_not_produce_separator(tmp_path, monkeypatch):
    """If global file is empty/whitespace, only project content returned, no separator."""
    monkeypatch.setenv("HOME", str(tmp_path))

    global_cody_dir = tmp_path / ".cody"
    global_cody_dir.mkdir()
    (global_cody_dir / CODY_MD_FILENAME).write_text("   ", encoding="utf-8")

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    (project_dir / CODY_MD_FILENAME).write_text("Project only", encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result == "Project only"
    assert "---" not in result


# ---------------------------------------------------------------------------
# load_project_instructions — unicode / encoding
# ---------------------------------------------------------------------------


def test_handles_unicode_content(tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))

    project_dir = tmp_path / "project"
    project_dir.mkdir()
    content = "# 项目说明\n\n这是一个 Python 项目。"
    (project_dir / CODY_MD_FILENAME).write_text(content, encoding="utf-8")

    result = load_project_instructions(project_dir)
    assert result == content



"""LLM API retry with exponential backoff.

Wraps async callables to retry on transient LLM provider errors
(rate limit 429, server errors 5xx). Does NOT retry on client errors
(bad request, auth failure) or context overflow.
"""

import asyncio
import logging
import re
from dataclasses import dataclass
from typing import TypeVar

logger = logging.getLogger(__name__)

T = TypeVar("T")

# HTTP status codes that warrant a retry.
_RETRYABLE_STATUS_CODES = {429, 500, 502, 503, 529}

# Keywords in exception messages that indicate a transient failure.
_RETRYABLE_KEYWORDS = (
    "rate limit",
    "overloaded",
    "too many requests",
    "server error",
    "connection",
    "timed out",
    "timeout",
    "service unavailable",
    "bad gateway",
    "internal server error",
)

# Keywords that indicate non-retryable errors (fail fast).
_NON_RETRYABLE_KEYWORDS = (
    "context_length_exceeded",
    "context window",
    "maximum context",
    "invalid_api_key",
    "authentication",
    "permission",
    "invalid_request",
)


@dataclass
class RetryParams:
    """Runtime parameters for LLM API retry behaviour.

    Separate from ``config.RetryConfig`` (Pydantic model for YAML/JSON config).
    ``AgentRunner._retry_config()`` converts one to the other.
    """
    max_retries: int = 3
    base_delay: float = 2.0
    max_delay: float = 30.0
    enabled: bool = True


def is_retryable(exc: Exception) -> bool:
    """Determine whether an exception represents a transient failure.

    Returns ``True`` for rate-limit (429) and server errors (5xx).
    Returns ``False`` for client errors, auth failures, and context overflow.
    """
    msg = str(exc).lower()

    # Fast reject: non-retryable keywords take precedence.
    if any(kw in msg for kw in _NON_RETRYABLE_KEYWORDS):
        return False

    # Check for retryable HTTP status codes in the message.
    # Use word boundary to avoid false positives (e.g. "port 5029" matching 502).
    for code in _RETRYABLE_STATUS_CODES:
        if re.search(rf'\b{code}\b', msg):
            return True

    # Check for retryable keywords.
    return any(kw in msg for kw in _RETRYABLE_KEYWORDS)


async def with_retry(
    fn,
    *args,
    retry_config: RetryParams | None = None,
    **kwargs,
):
    """Call *fn* with exponential backoff on transient LLM errors.

    Only retries when :func:`is_retryable` returns ``True``.
    Raises the original exception immediately for non-retryable errors.
    """
    cfg = retry_config or RetryParams()
    if not cfg.enabled:
        return await fn(*args, **kwargs)

    last_exc: Exception | None = None
    for attempt in range(cfg.max_retries + 1):
        try:
            return await fn(*args, **kwargs)
        except Exception as exc:
            last_exc = exc
            if attempt >= cfg.max_retries or not is_retryable(exc):
                raise
            delay = min(cfg.base_delay * (2 ** attempt), cfg.max_delay)
            logger.warning(
                "LLM call failed (attempt %d/%d), retrying in %.1fs: %s",
                attempt + 1, cfg.max_retries, delay, exc,
            )
            await asyncio.sleep(delay)
    raise last_exc  # type: ignore[misc]  # unreachable, satisfies type checker


def with_retry_sync(
    fn,
    *args,
    retry_config: RetryParams | None = None,
    **kwargs,
):
    """Synchronous version of :func:`with_retry`.

    Uses ``time.sleep`` instead of ``asyncio.sleep``.
    """
    import time

    cfg = retry_config or RetryParams()
    if not cfg.enabled:
        return fn(*args, **kwargs)

    last_exc: Exception | None = None
    for attempt in range(cfg.max_retries + 1):
        try:
            return fn(*args, **kwargs)
        except Exception as exc:
            last_exc = exc
            if attempt >= cfg.max_retries or not is_retryable(exc):
                raise
            delay = min(cfg.base_delay * (2 ** attempt), cfg.max_delay)
            logger.warning(
                "LLM call failed (attempt %d/%d), retrying in %.1fs: %s",
                attempt + 1, cfg.max_retries, delay, exc,
            )
            time.sleep(delay)
    raise last_exc  # type: ignore[misc]  # unreachable, satisfies type checker



"""Tests for configuration management"""

import json
from pathlib import Path

import pytest

from cody.core.config import (
    AuthConfig,
    Config,
    MCPConfig,
    MCPServerConfig,
    SecurityConfig,
    SkillConfig,
)

# Env vars that Config._apply_env_overrides reads.
_CODY_ENV_VARS = [
    "CODY_MODEL",
    "CODY_MODEL_BASE_URL",
    "CODY_MODEL_API_KEY",
    "CODY_CODING_PLAN_KEY",
    "CODY_ENABLE_THINKING",
    "CODY_THINKING_BUDGET",
    "CODY_SMALL_MODEL",
    "CODY_SMALL_MODEL_BASE_URL",
    "CODY_SMALL_MODEL_API_KEY",
]


@pytest.fixture(autouse=True)
def _clean_cody_env(monkeypatch):
    """Ensure CODY_* env vars don't leak into config tests."""
    for var in _CODY_ENV_VARS:
        monkeypatch.delenv(var, raising=False)


# ── Default config ───────────────────────────────────────────────────────────


def test_default_config():
    config = Config()
    assert config.model == ""
    assert config.model_base_url is None
    assert config.model_api_key is None
    assert config.auth.type == "api_key"
    assert config.skills.enabled == []
    assert config.skills.disabled == []
    assert config.mcp.servers == []
    assert config.security.allowed_commands is None
    assert config.security.restricted_paths == []
    assert config.security.allowed_roots == []
    assert config.security.require_confirmation is True


# ── AuthConfig ───────────────────────────────────────────────────────────────


def test_auth_config_defaults():
    auth = AuthConfig()
    assert auth.type == "api_key"
    assert auth.token is None
    assert auth.api_key is None
    assert auth.refresh_token is None
    assert auth.expires_at is None


def test_auth_config_oauth():
    auth = AuthConfig(type="oauth", token="tok123", refresh_token="ref456")
    assert auth.type == "oauth"
    assert auth.token == "tok123"
    assert auth.refresh_token == "ref456"


# ── SkillConfig ──────────────────────────────────────────────────────────────


def test_skill_config_defaults():
    sc = SkillConfig()
    assert sc.enabled == []
    assert sc.disabled == []


def test_skill_config_with_values():
    sc = SkillConfig(enabled=["git", "docker"], disabled=["web"])
    assert "git" in sc.enabled
    assert "web" in sc.disabled


# ── MCPConfig ────────────────────────────────────────────────────────────────


def test_mcp_config_defaults():
    mcp = MCPConfig()
    assert mcp.servers == []


def test_mcp_server_config():
    server = MCPServerConfig(
        name="github",
        command="npx",
        args=["-y", "@modelcontextprotocol/server-github"],
        env={"GITHUB_TOKEN": "test"},
    )
    assert server.name == "github"
    assert server.command == "npx"
    assert len(server.args) == 2
    assert server.env["GITHUB_TOKEN"] == "test"


# ── SecurityConfig ───────────────────────────────────────────────────────────


def test_security_config_defaults():
    sec = SecurityConfig()
    assert sec.allowed_commands is None
    assert sec.restricted_paths == []
    assert sec.require_confirmation is True


def test_security_config_with_whitelist():
    sec = SecurityConfig(allowed_commands=["git", "npm"], restricted_paths=["/etc"])
    assert sec.allowed_commands == ["git", "npm"]
    assert sec.restricted_paths == ["/etc"]


# ── Config.load / Config.save ────────────────────────────────────────────────


def test_config_load_nonexistent():
    """Loading from nonexistent path returns defaults"""
    config = Config.load("/tmp/nonexistent_cody_config_12345.json")
    assert config.model == ""


def test_config_save_and_load(tmp_path):
    """Save config and load it back"""
    config = Config(model="test:model")
    config_path = tmp_path / "config.json"
    config.save(config_path)

    loaded = Config.load(config_path)
    assert loaded.model == "test:model"


def test_config_load_from_json_file(tmp_path):
    """Load config from a manually-written JSON file"""
    data = {
        "model": "openai:gpt-4",
        "skills": {"enabled": ["git"], "disabled": ["docker"]},
        "security": {"allowed_commands": ["git", "ls"]},
    }
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))

    config = Config.load(config_path)
    assert config.model == "openai:gpt-4"
    assert config.skills.enabled == ["git"]
    assert config.skills.disabled == ["docker"]
    assert config.security.allowed_commands == ["git", "ls"]


def test_config_save_creates_parent_dirs(tmp_path):
    """Save should create parent directories"""
    config = Config()
    config_path = tmp_path / "deep" / "nested" / "config.json"
    config.save(config_path)
    assert config_path.exists()


def test_config_load_default_fallback(tmp_path, monkeypatch):
    """Config.load(workdir=...) falls back to default when no config files exist"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    config = Config.load(workdir=tmp_path / "project")
    assert config.model == ""


# ── Config round-trip with nested structures ─────────────────────────────────


def test_config_roundtrip_with_mcp(tmp_path):
    config = Config(
        mcp=MCPConfig(servers=[
            MCPServerConfig(name="test", command="echo", args=["hello"]),
        ])
    )
    config_path = tmp_path / "config.json"
    config.save(config_path)

    loaded = Config.load(config_path)
    assert len(loaded.mcp.servers) == 1
    assert loaded.mcp.servers[0].name == "test"
    assert loaded.mcp.servers[0].command == "echo"


# ── Custom model provider fields ────────────────────────────────────────────


def test_config_with_custom_model_provider():
    """Config accepts model_base_url and model_api_key"""
    config = Config(
        model="glm-4",
        model_base_url="https://open.bigmodel.cn/api/paas/v4/",
        model_api_key="sk-test-key",
    )
    assert config.model == "glm-4"
    assert config.model_base_url == "https://open.bigmodel.cn/api/paas/v4/"
    assert config.model_api_key == "sk-test-key"


def test_config_load_with_custom_model(tmp_path):
    """Load config with custom model provider from JSON"""
    data = {
        "model": "qwen-coder-plus",
        "model_base_url": "https://dashscope.aliyuncs.com/compatible-mode/v1",
    }
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))

    config = Config.load(config_path)
    assert config.model == "qwen-coder-plus"
    assert config.model_base_url == "https://dashscope.aliyuncs.com/compatible-mode/v1"
    assert config.model_api_key is None


def test_config_save_excludes_api_key(tmp_path):
    """Save should NOT persist model_api_key to disk (security)."""
    config = Config(
        model="glm-4",
        model_base_url="https://open.bigmodel.cn/api/paas/v4/",
        model_api_key="sk-secret-key",
    )
    config_path = tmp_path / "config.json"
    config.save(config_path)

    saved_data = json.loads(config_path.read_text())
    assert "model_api_key" not in saved_data
    assert saved_data["model"] == "glm-4"
    assert saved_data["model_base_url"] == "https://open.bigmodel.cn/api/paas/v4/"


# ── Environment variable overrides ──────────────────────────────────────────


def test_config_env_overrides_model(tmp_path, monkeypatch):
    """CODY_MODEL env var overrides config file value"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_MODEL", "glm-4")
    config = Config.load(workdir=tmp_path / "project")
    assert config.model == "glm-4"


def test_config_env_overrides_base_url(tmp_path, monkeypatch):
    """CODY_MODEL_BASE_URL env var overrides config file value"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_MODEL_BASE_URL", "https://custom.api.com/v1")
    config = Config.load(workdir=tmp_path / "project")
    assert config.model_base_url == "https://custom.api.com/v1"


def test_config_env_overrides_api_key(tmp_path, monkeypatch):
    """CODY_MODEL_API_KEY env var overrides config file value"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_MODEL_API_KEY", "sk-from-env")
    config = Config.load(workdir=tmp_path / "project")
    assert config.model_api_key == "sk-from-env"


def test_config_env_overrides_file_values(tmp_path, monkeypatch):
    """Env vars take priority over config file values"""
    data = {
        "model": "glm-4",
        "model_base_url": "https://from-file.com/v1",
    }
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))

    monkeypatch.setenv("CODY_MODEL", "qwen-coder-plus")
    monkeypatch.setenv("CODY_MODEL_BASE_URL", "https://from-env.com/v1")
    monkeypatch.setenv("CODY_MODEL_API_KEY", "sk-env-key")

    config = Config.load(config_path)
    assert config.model == "qwen-coder-plus"
    assert config.model_base_url == "https://from-env.com/v1"
    assert config.model_api_key == "sk-env-key"


# ── Config.is_ready / Config.missing_fields ──────────────────────────────


def test_config_is_ready_with_model_and_base_url():
    """Config with model + base_url is ready"""
    config = Config(model="test-model", model_base_url="https://api.example.com/v1")
    assert config.is_ready() is True


def test_config_not_ready_no_model():
    """Config with no model is NOT ready"""
    config = Config(model_base_url="https://api.example.com/v1")
    assert config.is_ready() is False


def test_config_not_ready_no_base_url():
    """Config with no base_url is NOT ready"""
    config = Config(model="test-model")
    assert config.is_ready() is False


def test_config_not_ready_empty():
    """Default config is NOT ready"""
    config = Config()
    assert config.is_ready() is False


def test_config_missing_fields_no_config():
    """missing_fields reports missing model and base_url when not configured"""
    config = Config()
    missing = config.missing_fields()
    assert len(missing) == 2
    assert any("model" in m for m in missing)
    assert any("model_base_url" in m for m in missing)


def test_config_missing_fields_empty_when_ready():
    """missing_fields returns empty list when config is ready"""
    config = Config(model="test-model", model_base_url="https://api.example.com/v1")
    assert config.missing_fields() == []


def test_config_load_strips_legacy_oauth(tmp_path):
    """Config.load ignores legacy claude_oauth_token in JSON files"""
    data = {"model": "test-model", "claude_oauth_token": "old-oauth-token"}
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))

    config = Config.load(config_path)
    assert config.model == "test-model"
    assert not hasattr(config, "claude_oauth_token") or getattr(config, "claude_oauth_token", None) is None


# ── Legacy CODY_CODING_PLAN_KEY env var compat ───────────────────────────────


def test_config_env_coding_plan_key_maps_to_model_api_key(tmp_path, monkeypatch):
    """CODY_CODING_PLAN_KEY env var maps to model_api_key"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_CODING_PLAN_KEY", "sk-sp-from-env")
    config = Config.load(workdir=tmp_path / "project")
    assert config.model_api_key == "sk-sp-from-env"


def test_config_env_coding_plan_key_does_not_override_model_api_key(tmp_path, monkeypatch):
    """CODY_CODING_PLAN_KEY does NOT override explicit CODY_MODEL_API_KEY"""
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_MODEL_API_KEY", "sk-explicit")
    monkeypatch.setenv("CODY_CODING_PLAN_KEY", "sk-sp-from-env")
    config = Config.load(workdir=tmp_path / "project")
    assert config.model_api_key == "sk-explicit"


def test_config_load_strips_legacy_coding_plan_fields(tmp_path):
    """Config.load ignores legacy coding_plan fields in JSON files"""
    data = {"model": "qwen3.5", "coding_plan_key": "sk-sp-old", "coding_plan_protocol": "anthropic"}
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))

    config = Config.load(config_path)
    assert config.model == "qwen3.5"
    assert not hasattr(config, "coding_plan_key") or getattr(config, "coding_plan_key", None) is None


# ── Config.load with workdir ────────────────────────────────────────────────


def test_config_load_workdir_finds_project_config(tmp_path, monkeypatch):
    """Config.load(workdir=X) looks for .cody/config.json in X, not cwd."""
    # Put a project config in a different directory than cwd
    project_dir = tmp_path / "my-project"
    (project_dir / ".cody").mkdir(parents=True)
    (project_dir / ".cody" / "config.json").write_text(
        json.dumps({"model": "project-model"})
    )

    # cwd has no config
    cwd_dir = tmp_path / "elsewhere"
    cwd_dir.mkdir()
    monkeypatch.setattr(Path, "cwd", lambda: cwd_dir)
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "nohome")
    (tmp_path / "nohome").mkdir()

    config = Config.load(workdir=project_dir)
    assert config.model == "project-model"


def test_config_load_defaults_workdir_to_cwd(tmp_path, monkeypatch):
    """Config.load() without path or workdir defaults workdir to cwd."""
    monkeypatch.chdir(tmp_path)
    config = Config.load()
    assert isinstance(config, Config)


def test_config_load_workdir_no_project_config_falls_to_global(tmp_path, monkeypatch):
    """When workdir has no .cody/config.json, global config is tried."""
    project_dir = tmp_path / "project"
    project_dir.mkdir()

    home_dir = tmp_path / "home"
    (home_dir / ".cody").mkdir(parents=True)
    (home_dir / ".cody" / "config.json").write_text(
        json.dumps({"model": "global-model"})
    )

    monkeypatch.setattr(Path, "home", lambda: home_dir)

    config = Config.load(workdir=project_dir)
    assert config.model == "global-model"


# ── SecurityConfig.allowed_roots ─────────────────────────────────────────────


def test_security_config_allowed_roots_default():
    sec = SecurityConfig()
    assert sec.allowed_roots == []


def test_security_config_allowed_roots_from_json(tmp_path):
    data = {"security": {"allowed_roots": ["/tmp/shared", "/data/models"]}}
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))
    config = Config.load(config_path)
    assert config.security.allowed_roots == ["/tmp/shared", "/data/models"]


def test_apply_overrides_extra_roots_additive():
    config = Config()
    config.security.allowed_roots = ["/existing/root"]
    config.apply_overrides(extra_roots=["/new/root"])
    assert "/existing/root" in config.security.allowed_roots
    assert "/new/root" in config.security.allowed_roots
    assert len(config.security.allowed_roots) == 2


def test_apply_overrides_extra_roots_no_duplicates():
    config = Config()
    config.security.allowed_roots = ["/some/root"]
    config.apply_overrides(extra_roots=["/some/root"])
    assert config.security.allowed_roots.count("/some/root") == 1


def test_apply_overrides_extra_roots_none_is_noop():
    config = Config()
    config.security.allowed_roots = ["/a"]
    config.apply_overrides(extra_roots=None)
    assert config.security.allowed_roots == ["/a"]


def test_apply_overrides_extra_roots_empty_is_noop():
    config = Config()
    config.security.allowed_roots = ["/a"]
    config.apply_overrides(extra_roots=[])
    assert config.security.allowed_roots == ["/a"]


# ── Small model configuration ────────────────────────────────────────────────


def test_small_model_defaults():
    config = Config()
    assert config.small_model is None
    assert config.small_model_base_url is None
    assert config.small_model_api_key is None


def test_small_model_from_json(tmp_path):
    data = {
        "model": "claude-opus",
        "small_model": "gpt-4o-mini",
        "small_model_base_url": "https://api.openai.com/v1",
    }
    config_path = tmp_path / "config.json"
    config_path.write_text(json.dumps(data))
    config = Config.load(config_path)
    assert config.small_model == "gpt-4o-mini"
    assert config.small_model_base_url == "https://api.openai.com/v1"


def test_small_model_env_overrides(tmp_path, monkeypatch):
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    (tmp_path / "project").mkdir()
    (tmp_path / "home").mkdir()

    monkeypatch.setenv("CODY_SMALL_MODEL", "gpt-4o-mini")
    monkeypatch.setenv("CODY_SMALL_MODEL_BASE_URL", "https://api.openai.com/v1")
    monkeypatch.setenv("CODY_SMALL_MODEL_API_KEY", "sk-small-key")
    config = Config.load(workdir=tmp_path / "project")
    assert config.small_model == "gpt-4o-mini"
    assert config.small_model_base_url == "https://api.openai.com/v1"
    assert config.small_model_api_key == "sk-small-key"


def test_small_model_api_key_excluded_from_save(tmp_path):
    config = Config(
        model="test",
        small_model="gpt-4o-mini",
        small_model_api_key="sk-secret",
    )
    config_path = tmp_path / "config.json"
    config.save(config_path)

    saved_data = json.loads(config_path.read_text())
    assert "small_model_api_key" not in saved_data
    assert saved_data["small_model"] == "gpt-4o-mini"



"""Tests for core error types (no web/FastAPI dependency)."""

from cody.core.errors import CodyAPIError, ErrorCode, ErrorDetail


# ── ErrorDetail model ───────────────────────────────────────────────────────


def test_error_detail_model():
    detail = ErrorDetail(code=ErrorCode.TOOL_NOT_FOUND, message="not found")
    assert detail.code == ErrorCode.TOOL_NOT_FOUND
    assert detail.message == "not found"
    assert detail.details is None


def test_error_detail_with_extra():
    detail = ErrorDetail(
        code=ErrorCode.TOOL_ERROR,
        message="failed",
        details={"tool": "grep"},
    )
    assert detail.details["tool"] == "grep"


# ── CodyAPIError ────────────────────────────────────────────────────────────


def test_cody_api_error_to_detail():
    err = CodyAPIError(
        code=ErrorCode.SESSION_NOT_FOUND,
        message="Session not found: abc",
        status_code=404,
    )
    body = err.to_detail()
    assert body["error"]["code"] == "SESSION_NOT_FOUND"
    assert body["error"]["message"] == "Session not found: abc"


def test_cody_api_error_with_details():
    err = CodyAPIError(
        code=ErrorCode.TOOL_ERROR,
        message="failed",
        status_code=500,
        details={"tool": "grep", "reason": "timeout"},
    )
    body = err.to_detail()
    assert body["error"]["details"]["tool"] == "grep"
    assert body["error"]["details"]["reason"] == "timeout"


# ── ErrorCode enum ──────────────────────────────────────────────────────────


def test_error_code_values():
    assert ErrorCode.INVALID_PARAMS == "INVALID_PARAMS"
    assert ErrorCode.TOOL_NOT_FOUND == "TOOL_NOT_FOUND"
    assert ErrorCode.SESSION_NOT_FOUND == "SESSION_NOT_FOUND"
    assert ErrorCode.SERVER_ERROR == "SERVER_ERROR"
    assert ErrorCode.MCP_ERROR == "MCP_ERROR"
    assert ErrorCode.AGENT_ERROR == "AGENT_ERROR"
    assert ErrorCode.AGENT_NOT_FOUND == "AGENT_NOT_FOUND"
    assert ErrorCode.AGENT_LIMIT_REACHED == "AGENT_LIMIT_REACHED"



"""CLI configuration commands."""

from pathlib import Path

import click

from ...core import Config
from ...shared import resolve_config_path
from ..utils import console, _mask_api_key, _interactive_setup


@click.group()
def config():
    """Manage configuration"""


@config.command('show')
def config_show():
    """Show current configuration"""
    cfg = Config.load(workdir=Path.cwd())
    data = cfg.model_dump(exclude_none=True)
    # Mask sensitive fields for display
    if "model_api_key" in data:
        data["model_api_key"] = _mask_api_key(data["model_api_key"])
    if "auth" in data:
        for key in ("api_key", "token", "refresh_token"):
            if key in data["auth"]:
                data["auth"][key] = _mask_api_key(data["auth"][key])
    import json as _json
    console.print_json(_json.dumps(data, indent=2, default=str))


@config.command('set')
@click.argument('key')
@click.argument('value')
def config_set(key, value):
    """Set configuration value

    Supported keys: model, model_base_url, model_api_key, enable_thinking, thinking_budget
    """
    cfg = Config.load(workdir=Path.cwd())

    if key == 'model':
        cfg.model = value
    elif key == 'model_base_url':
        cfg.model_base_url = value
    elif key == 'model_api_key':
        cfg.model_api_key = value
    elif key == 'enable_thinking':
        cfg.enable_thinking = value.lower() in ("1", "true", "yes")
    elif key == 'thinking_budget':
        cfg.thinking_budget = int(value)
    else:
        console.print(f"[yellow]Unknown config key: {key}[/yellow]")
        console.print("[dim]Supported: model, model_base_url, model_api_key, enable_thinking, thinking_budget[/dim]")
        return

    cfg.save(resolve_config_path())
    display_value = _mask_api_key(value) if "key" in key else value
    console.print(f"[green]Set {key} = {display_value}[/green]")


@config.command('setup')
def config_setup():
    """Interactive configuration wizard

    Set up your AI model provider, API key, and preferences.
    Configuration is saved to ~/.cody/config.json.
    """
    _interactive_setup()



"""Tests for ProjectMemoryStore."""


import pytest

from cody.core.memory import (
    MAX_ENTRIES_PER_CATEGORY,
    MemoryEntry,
    ProjectMemoryStore,
    _project_id,
)


class TestProjectId:
    def test_deterministic(self, tmp_path):
        p1 = _project_id(tmp_path)
        p2 = _project_id(tmp_path)
        assert p1 == p2

    def test_different_paths(self, tmp_path):
        d1 = tmp_path / "project-a"
        d2 = tmp_path / "project-b"
        d1.mkdir()
        d2.mkdir()
        assert _project_id(d1) != _project_id(d2)

    def test_length(self, tmp_path):
        assert len(_project_id(tmp_path)) == 12


class TestProjectMemoryStore:
    @pytest.fixture
    def store(self, tmp_path):
        return ProjectMemoryStore("test_proj", base_dir=tmp_path)

    @pytest.fixture
    def store_from_workdir(self, tmp_path):
        workdir = tmp_path / "myproject"
        workdir.mkdir()
        base = tmp_path / "memory"
        return ProjectMemoryStore.from_workdir(workdir, base_dir=base)

    def test_from_workdir(self, store_from_workdir):
        assert store_from_workdir.store_dir.exists()

    @pytest.mark.asyncio
    async def test_add_and_get_entries(self, store):
        entries = [
            MemoryEntry(content="Use ruff for linting"),
            MemoryEntry(content="Line width is 100"),
        ]
        await store.add_entries("conventions", entries)

        all_entries = store.get_all_entries()
        assert "conventions" in all_entries
        assert len(all_entries["conventions"]) == 2
        assert all_entries["conventions"][0].content == "Use ruff for linting"

    @pytest.mark.asyncio
    async def test_invalid_category(self, store):
        with pytest.raises(ValueError, match="Unknown memory category"):
            await store.add_entries("invalid_cat", [MemoryEntry(content="x")])

    @pytest.mark.asyncio
    async def test_max_entries_enforced(self, store):
        entries = [MemoryEntry(content=f"entry-{i}") for i in range(60)]
        await store.add_entries("patterns", entries)

        all_entries = store.get_all_entries()
        assert len(all_entries["patterns"]) == MAX_ENTRIES_PER_CATEGORY
        # Should keep the most recent
        assert all_entries["patterns"][-1].content == "entry-59"

    @pytest.mark.asyncio
    async def test_persistence(self, tmp_path):
        store1 = ProjectMemoryStore("persist_test", base_dir=tmp_path)
        await store1.add_entries("issues", [MemoryEntry(content="Bug in parser")])

        store2 = ProjectMemoryStore("persist_test", base_dir=tmp_path)
        all_entries = store2.get_all_entries()
        assert len(all_entries["issues"]) == 1
        assert all_entries["issues"][0].content == "Bug in parser"

    def test_get_memory_for_prompt_empty(self, store):
        prompt = store.get_memory_for_prompt()
        assert prompt == ""

    @pytest.mark.asyncio
    async def test_get_memory_for_prompt_with_entries(self, store):
        await store.add_entries("conventions", [MemoryEntry(content="Use snake_case")])
        await store.add_entries("decisions", [MemoryEntry(content="Chose FastAPI")])

        prompt = store.get_memory_for_prompt()
        assert "Project Memory" in prompt
        assert "Use snake_case" in prompt
        assert "Chose FastAPI" in prompt

    @pytest.mark.asyncio
    async def test_cleanup_low_confidence(self, store):
        entries = [
            MemoryEntry(content="keep me", confidence=0.8),
            MemoryEntry(content="remove me", confidence=0.1),
        ]
        await store.add_entries("issues", entries)
        await store.cleanup()

        all_entries = store.get_all_entries()
        assert len(all_entries["issues"]) == 1
        assert all_entries["issues"][0].content == "keep me"

    def test_count_empty(self, store):
        counts = store.count()
        assert all(c == 0 for c in counts.values())

    @pytest.mark.asyncio
    async def test_count_with_entries(self, store):
        await store.add_entries("conventions", [MemoryEntry(content="x")])
        await store.add_entries("patterns", [
            MemoryEntry(content="a"),
            MemoryEntry(content="b"),
        ])
        counts = store.count()
        assert counts["conventions"] == 1
        assert counts["patterns"] == 2
        assert counts["issues"] == 0

    @pytest.mark.asyncio
    async def test_clear(self, store):
        await store.add_entries("conventions", [MemoryEntry(content="x")])
        await store.add_entries("patterns", [MemoryEntry(content="y")])
        store.clear()

        counts = store.count()
        assert all(c == 0 for c in counts.values())

    @pytest.mark.asyncio
    async def test_memory_entry_fields(self, store):
        entry = MemoryEntry(
            content="Always run tests before commit",
            source_task_id="task_001",
            source_task_title="Fix CI",
            confidence=0.95,
            tags=["ci", "testing"],
        )
        await store.add_entries("conventions", [entry])

        loaded = store.get_all_entries()["conventions"][0]
        assert loaded.content == "Always run tests before commit"
        assert loaded.source_task_id == "task_001"
        assert loaded.confidence == 0.95
        assert "ci" in loaded.tags

    @pytest.mark.asyncio
    async def test_low_confidence_excluded_from_prompt(self, store):
        entries = [
            MemoryEntry(content="visible", confidence=0.5),
            MemoryEntry(content="hidden", confidence=0.1),
        ]
        await store.add_entries("conventions", entries)
        prompt = store.get_memory_for_prompt()
        assert "visible" in prompt
        assert "hidden" not in prompt



"""Tests for step hook / middleware (#11)."""

import pytest
from unittest.mock import MagicMock, patch

from pydantic_ai import RunContext

from cody.core.config import Config
from cody.core.deps import CodyDeps
from cody.core.runner import AgentRunner
from cody.core.skill_manager import SkillManager
from cody.core.tools._base import _with_model_retry


# ── Sample tool for testing ──────────────────────────────────────────────────


async def sample_tool(ctx: RunContext[CodyDeps], query: str) -> str:
    """A sample tool that echoes the query."""
    return f"result:{query}"


# ── Mock context helper ──────────────────────────────────────────────────────


class _MockCtx:
    def __init__(self, deps):
        self.deps = deps


def _make_deps(
    tmp_path,
    before_hooks=None,
    after_hooks=None,
) -> CodyDeps:
    config = Config()
    return CodyDeps(
        config=config,
        workdir=tmp_path,
        skill_manager=SkillManager(config, workdir=tmp_path),
        before_tool_hooks=before_hooks or [],
        after_tool_hooks=after_hooks or [],
    )


# ── _with_model_retry hook integration ───────────────────────────────────────


class TestBeforeToolHook:
    @pytest.mark.asyncio
    async def test_before_hook_called(self, tmp_path):
        """before_tool hook is called with tool name and args."""
        calls = []

        async def hook(tool_name, args):
            calls.append((tool_name, args))
            return args  # proceed

        deps = _make_deps(tmp_path, before_hooks=[hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"
        assert calls == [("sample_tool", {"query": "hello"})]

    @pytest.mark.asyncio
    async def test_before_hook_modifies_args(self, tmp_path):
        """before_tool hook can modify args."""
        async def hook(tool_name, args):
            args["query"] = "modified"
            return args

        deps = _make_deps(tmp_path, before_hooks=[hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="original")
        assert result == "result:modified"

    @pytest.mark.asyncio
    async def test_before_hook_rejects(self, tmp_path):
        """before_tool hook returning None rejects the call."""
        from pydantic_ai import ModelRetry

        async def hook(tool_name, args):
            return None  # reject

        deps = _make_deps(tmp_path, before_hooks=[hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        with pytest.raises(ModelRetry, match="rejected"):
            await wrapped(ctx, query="hello")

    @pytest.mark.asyncio
    async def test_multiple_before_hooks(self, tmp_path):
        """Multiple before hooks run in order."""
        order = []

        async def hook1(tool_name, args):
            order.append("hook1")
            args["query"] = args["query"] + "_1"
            return args

        async def hook2(tool_name, args):
            order.append("hook2")
            args["query"] = args["query"] + "_2"
            return args

        deps = _make_deps(tmp_path, before_hooks=[hook1, hook2])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="x")
        assert result == "result:x_1_2"
        assert order == ["hook1", "hook2"]


class TestBeforeToolHookErrorHandling:
    @pytest.mark.asyncio
    async def test_before_hook_exception_skipped(self, tmp_path):
        """A before_tool hook that raises is skipped, tool still runs."""
        async def bad_hook(tool_name, args):
            raise RuntimeError("hook crashed")

        deps = _make_deps(tmp_path, before_hooks=[bad_hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"

    @pytest.mark.asyncio
    async def test_before_hook_invalid_return_type_skipped(self, tmp_path):
        """A before_tool hook returning non-dict is skipped."""
        async def bad_hook(tool_name, args):
            return 42  # not a dict

        deps = _make_deps(tmp_path, before_hooks=[bad_hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"


class TestAfterToolHookErrorHandling:
    @pytest.mark.asyncio
    async def test_after_hook_exception_skipped(self, tmp_path):
        """An after_tool hook that raises is skipped, result preserved."""
        async def bad_hook(tool_name, args, result):
            raise RuntimeError("hook crashed")

        deps = _make_deps(tmp_path, after_hooks=[bad_hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"

    @pytest.mark.asyncio
    async def test_after_hook_invalid_return_type_skipped(self, tmp_path):
        """An after_tool hook returning non-str is skipped."""
        async def bad_hook(tool_name, args, result):
            return 42  # not a str

        deps = _make_deps(tmp_path, after_hooks=[bad_hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"


class TestAfterToolHook:
    @pytest.mark.asyncio
    async def test_after_hook_called(self, tmp_path):
        """after_tool hook is called with tool name, args, and result."""
        calls = []

        async def hook(tool_name, args, result):
            calls.append((tool_name, args, result))
            return result

        deps = _make_deps(tmp_path, after_hooks=[hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="test")
        assert result == "result:test"
        assert len(calls) == 1
        assert calls[0][0] == "sample_tool"
        assert calls[0][2] == "result:test"

    @pytest.mark.asyncio
    async def test_after_hook_modifies_result(self, tmp_path):
        """after_tool hook can modify the result."""
        async def hook(tool_name, args, result):
            return result.upper()

        deps = _make_deps(tmp_path, after_hooks=[hook])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "RESULT:HELLO"

    @pytest.mark.asyncio
    async def test_multiple_after_hooks(self, tmp_path):
        """Multiple after hooks chain results."""
        async def hook1(tool_name, args, result):
            return result + " [hook1]"

        async def hook2(tool_name, args, result):
            return result + " [hook2]"

        deps = _make_deps(tmp_path, after_hooks=[hook1, hook2])
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="x")
        assert result == "result:x [hook1] [hook2]"


class TestNoHooks:
    @pytest.mark.asyncio
    async def test_no_hooks_normal_execution(self, tmp_path):
        """Without hooks, tool executes normally."""
        deps = _make_deps(tmp_path)
        ctx = _MockCtx(deps)
        wrapped = _with_model_retry(sample_tool)
        result = await wrapped(ctx, query="hello")
        assert result == "result:hello"


# ── Runner-level tests ───────────────────────────────────────────────────────


class TestRunnerHooks:
    def test_runner_stores_hooks(self, tmp_path):
        """AgentRunner stores hooks."""
        async def hook(n, a):
            return a

        with patch.object(AgentRunner, "_create_agent", return_value=MagicMock()):
            runner = AgentRunner(
                config=Config(), workdir=tmp_path,
                before_tool_hooks=[hook],
                after_tool_hooks=[hook],
            )
        assert runner._before_tool_hooks == [hook]
        assert runner._after_tool_hooks == [hook]

    def test_runner_default_no_hooks(self, tmp_path):
        """Without hooks, lists are empty."""
        with patch.object(AgentRunner, "_create_agent", return_value=MagicMock()):
            runner = AgentRunner(config=Config(), workdir=tmp_path)
        assert runner._before_tool_hooks == []
        assert runner._after_tool_hooks == []


# ── SDK Builder-level tests ──────────────────────────────────────────────────


class TestBuilderHooks:
    def test_before_tool_method(self):
        from cody.sdk.client import CodyBuilder
        async def hook(n, a):
            return a
        builder = CodyBuilder()
        result = builder.before_tool(hook)
        assert result is builder
        assert hook in builder._before_tool_hooks

    def test_after_tool_method(self):
        from cody.sdk.client import CodyBuilder
        async def hook(n, a, r):
            return r
        builder = CodyBuilder()
        result = builder.after_tool(hook)
        assert result is builder
        assert hook in builder._after_tool_hooks

    def test_builder_passes_hooks_to_client(self):
        from cody.sdk.client import Cody, AsyncCodyClient
        async def bh(n, a):
            return a
        async def ah(n, a, r):
            return r
        with patch.object(AsyncCodyClient, "__init__", return_value=None) as mock_init:
            Cody().before_tool(bh).after_tool(ah).build()
            _, kwargs = mock_init.call_args
            assert kwargs["before_tool_hooks"] == [bh]
            assert kwargs["after_tool_hooks"] == [ah]



"""Cody agent dependencies - extracted to break circular imports."""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from .interaction import InteractionRequest, InteractionResponse

from .config import Config
from .lsp_client import LSPClient
from .mcp_client import MCPClient
from .permissions import PermissionManager
from .skill_manager import SkillManager
from .storage import AuditLoggerProtocol, FileHistoryProtocol, MemoryStoreProtocol
from .sub_agent import SubAgentManager


class _UnsetType:
    """Sentinel type to distinguish 'not passed' from explicit ``None``.

    Use ``UNSET`` (the singleton instance) as default parameter value
    when ``None`` carries meaning (e.g. "disable this feature").
    """
    _instance: "_UnsetType | None" = None

    def __new__(cls) -> "_UnsetType":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return "UNSET"

    def __bool__(self) -> bool:
        return False


UNSET = _UnsetType()
"""Shared sentinel — use ``is UNSET`` to check."""


# Hook type aliases
if TYPE_CHECKING:
    BeforeToolHook = Callable[[str, dict], Awaitable[dict | None]]
    AfterToolHook = Callable[[str, dict, str], Awaitable[str]]
else:
    BeforeToolHook = Any
    AfterToolHook = Any


class ToolContext:
    """Lightweight RunContext stand-in for direct tool invocation."""
    def __init__(self, deps: 'CodyDeps'):
        self.deps = deps


@dataclass
class CodyDeps:
    """Dependencies for Cody Agent"""
    config: Config
    workdir: Path
    skill_manager: SkillManager
    allowed_roots: list[Path] = field(default_factory=list)
    strict_read_boundary: bool = False
    mcp_client: Optional[MCPClient] = None
    sub_agent_manager: Optional[SubAgentManager] = None
    lsp_client: Optional[LSPClient] = None
    audit_logger: Optional[AuditLoggerProtocol] = None
    permission_manager: Optional[PermissionManager] = None
    file_history: Optional[FileHistoryProtocol] = None
    todo_list: Optional[list] = None
    memory_store: Optional[MemoryStoreProtocol] = None
    interaction_handler: Optional[Callable[[InteractionRequest], Awaitable[InteractionResponse]]] = None
    before_tool_hooks: list[BeforeToolHook] = field(default_factory=list)
    after_tool_hooks: list[AfterToolHook] = field(default_factory=list)
    auto_approved_tools: set[str] = field(default_factory=set)



"""Cody SDK - Enhanced client with Builder pattern, events, and metrics.

This module wraps core directly (AgentRunner + SessionStore) and adds:
- Builder pattern construction
- Convenience methods for common operations
- Event hooks for monitoring
- Metrics collection
"""

import asyncio
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import AsyncIterator, Literal, Optional, overload

from .config import (
    CircuitBreakerConfig as SDKCircuitBreakerConfig,
    InteractionConfig as SDKInteractionConfig,
    MCPServerConfig as SDKMCPServerConfig,
    SDKConfig,
    config as make_config,
)
from .errors import (
    CodyConfigError,
    CodyNotFoundError,
    CodyToolError,
)
from .events import (
    ContextCompactEvent as SDKContextCompactEvent,
    EventManager,
    EventType,
    ModelEvent,
    RunEvent,
    SessionEvent,
    StreamEvent as SDKStreamEvent,
    ThinkingEvent as SDKThinkingEvent,
    ToolEvent,
)
from .metrics import MetricsCollector, TokenUsage
from .types import (
    RunResult,
    SessionDetail,
    SessionInfo,
    StreamChunk,
    ToolResult,
    _event_to_chunk,
    _usage_from_result,
)


from ..core.deps import UNSET, _UnsetType
from ..core.storage import (
    AuditLoggerProtocol,
    FileHistoryProtocol,
    MemoryStoreProtocol,
    SessionStoreProtocol,
)


# ── Builder Pattern ─────────────────────────────────────────────────────────


@dataclass
class CodyBuilder:
    """Builder for creating AsyncCodyClient instances.

    Usage:
        client = (
            Cody()
            .workdir("/path/to/project")
            .model("your-model-name")
            .base_url("https://api.example.com/v1")
            .build()
        )
    """

    _workdir: Optional[str] = None
    _model: Optional[str] = None
    _api_key: Optional[str] = None
    _base_url: Optional[str] = None
    _enable_thinking: bool = False
    _thinking_budget: Optional[int] = None
    _permissions: dict = field(default_factory=dict)
    _allowed_roots: list[str] = field(default_factory=list)
    _strict_read_boundary: bool = False
    _db_path: Optional[str] = None
    _enable_metrics: bool = False
    _enable_events: bool = False
    _mcp_servers: list[dict | SDKMCPServerConfig] = field(default_factory=list)
    _auto_start_mcp: bool = False
    _interaction: SDKInteractionConfig | None = None
    _circuit_breaker: SDKCircuitBreakerConfig | None = None
    _skill_dirs: list[str] = field(default_factory=list)
    _lsp_languages: list[str] = field(default_factory=lambda: ["python", "typescript", "go"])
    _event_handlers: list[tuple] = field(default_factory=list)
    _custom_tools: list = field(default_factory=list)
    _system_prompt: str | None = None
    _extra_system_prompt: str | None = None
    _before_tool_hooks: list = field(default_factory=list)
    _after_tool_hooks: list = field(default_factory=list)
    _session_store: SessionStoreProtocol | _UnsetType | None = UNSET
    _audit_logger: AuditLoggerProtocol | _UnsetType | None = UNSET
    _file_history: FileHistoryProtocol | _UnsetType | None = UNSET
    _memory_store: MemoryStoreProtocol | _UnsetType | None = UNSET
    _stateless: bool = False

    def workdir(self, path: str) -> "CodyBuilder":
        """Set working directory."""
        self._workdir = path
        return self

    def model(self, model: str) -> "CodyBuilder":
        """Set model name (e.g., 'claude-sonnet-4-0', 'gpt-4o')."""
        self._model = model
        return self

    def api_key(self, key: str) -> "CodyBuilder":
        """Set API key."""
        self._api_key = key
        return self

    def base_url(self, url: str) -> "CodyBuilder":
        """Set custom API base URL."""
        self._base_url = url
        return self

    def thinking(self, enabled: bool = True, budget: Optional[int] = None) -> "CodyBuilder":
        """Enable thinking mode with optional token budget."""
        self._enable_thinking = enabled
        if budget is not None:
            self._thinking_budget = budget
        return self

    def permission(self, tool: str, level: str) -> "CodyBuilder":
        """Set permission for a specific tool."""
        self._permissions[tool] = level
        return self

    def allowed_root(self, path: str) -> "CodyBuilder":
        """Add an allowed root path for file operations."""
        self._allowed_roots.append(path)
        return self

    def allowed_roots(self, paths: list[str]) -> "CodyBuilder":
        """Set multiple allowed root paths."""
        self._allowed_roots = paths
        return self

    def strict_read_boundary(self, enabled: bool = True) -> "CodyBuilder":
        """When True, read operations are also restricted to workdir + allowed_roots."""
        self._strict_read_boundary = enabled
        return self

    def skill_dir(self, path: str) -> "CodyBuilder":
        """Add a custom skill directory."""
        self._skill_dirs.append(path)
        return self

    def skill_dirs(self, paths: list[str]) -> "CodyBuilder":
        """Set custom skill directories."""
        self._skill_dirs = list(paths)
        return self

    def db_path(self, path: str) -> "CodyBuilder":
        """Set session database path."""
        self._db_path = path
        return self

    def enable_metrics(self) -> "CodyBuilder":
        """Enable metrics collection."""
        self._enable_metrics = True
        return self

    def enable_events(self) -> "CodyBuilder":
        """Enable event system."""
        self._enable_events = True
        return self

    def interaction(
        self,
        enabled: bool = True,
        timeout: float = 30.0,
    ) -> "CodyBuilder":
        """Configure human-in-the-loop interaction.

        When enabled, the runner pauses on interaction requests (e.g. the
        ``question`` tool) and waits for a human response via
        ``submit_interaction()``.  If no response arrives within *timeout*
        seconds, the run is terminated with an ``InteractionTimeoutEvent``.

        Only effective with async methods (``run()`` / ``stream()``).
        ``run_sync()`` always auto-approves regardless of this setting.

            Cody().interaction(enabled=True, timeout=30).build()
        """
        self._interaction = SDKInteractionConfig(enabled=enabled, timeout=timeout)
        return self

    def circuit_breaker(
        self,
        config: SDKCircuitBreakerConfig | None = None,
        *,
        enabled: bool = True,
        max_tokens: int = 200_000,
        max_cost_usd: float = 5.0,
        max_steps: int = 0,
        loop_detect_turns: int = 6,
        loop_similarity_threshold: float = 0.9,
        model_prices: dict[str, float] | None = None,
    ) -> "CodyBuilder":
        """Configure circuit breaker for automatic run termination.

        Accepts either a CircuitBreakerConfig object or keyword arguments:

            # Option A: config object
            Cody().circuit_breaker(CircuitBreakerConfig(max_cost_usd=10.0)).build()

            # Option B: keyword arguments
            Cody().circuit_breaker(max_cost_usd=10.0, max_steps=50).build()
        """
        if config is not None:
            self._circuit_breaker = config
        else:
            self._circuit_breaker = SDKCircuitBreakerConfig(
                enabled=enabled,
                max_tokens=max_tokens,
                max_cost_usd=max_cost_usd,
                max_steps=max_steps,
                loop_detect_turns=loop_detect_turns,
                loop_similarity_threshold=loop_similarity_threshold,
                model_prices=model_prices or {},
            )
        return self

    def mcp_server(self, server: dict | SDKMCPServerConfig) -> "CodyBuilder":
        """Add MCP server configuration (dict or MCPServerConfig)."""
        self._mcp_servers.append(server)
        return self

    def mcp_stdio_server(
        self,
        name: str,
        command: str,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
    ) -> "CodyBuilder":
        """Add a stdio-based MCP server (subprocess)."""
        self._mcp_servers.append(SDKMCPServerConfig(
            name=name, transport='stdio',
            command=command, args=args or [], env=env or {},
        ))
        return self

    def mcp_http_server(
        self,
        name: str,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> "CodyBuilder":
        """Add an HTTP-based MCP server (remote endpoint)."""
        self._mcp_servers.append(SDKMCPServerConfig(
            name=name, transport='http',
            url=url, headers=headers or {},
        ))
        return self

    def auto_start_mcp(self, enabled: bool = True) -> "CodyBuilder":
        """Auto-start MCP servers on first run(). Default is False."""
        self._auto_start_mcp = enabled
        return self

    def lsp_languages(self, languages: list[str]) -> "CodyBuilder":
        """Set LSP languages to enable."""
        self._lsp_languages = languages
        return self

    def tool(self, func) -> "CodyBuilder":
        """Register a custom tool function.

        The function must be an async callable with the signature::

            async def my_tool(ctx: RunContext[CodyDeps], arg: str) -> str:
                ...

        Custom tools are registered alongside built-in tools and are
        available to the agent during ``run()`` / ``stream()`` calls.

        Example::

            async def lookup_jira(ctx: RunContext[CodyDeps], ticket: str) -> str:
                \"\"\"Look up a Jira ticket by ID.\"\"\"
                return await fetch_jira(ticket)

            client = Cody().tool(lookup_jira).build()
        """
        self._custom_tools.append(func)
        return self

    def system_prompt(self, text: str) -> "CodyBuilder":
        """Replace the default base persona with a custom system prompt.

        The custom prompt replaces only the base persona. CODY.md project
        instructions, project memory, and skills are still appended.

        Example::

            client = (
                Cody()
                .system_prompt("You are a security-focused code review agent.")
                .build()
            )
        """
        self._system_prompt = text
        return self

    def extra_system_prompt(self, text: str) -> "CodyBuilder":
        """Append additional instructions after all built-in system prompt parts.

        Unlike ``system_prompt()``, this does not replace the default persona
        — it adds to it.  Use this for injecting business context or
        run-specific instructions.

        Example::

            client = (
                Cody()
                .extra_system_prompt("Always respond in Chinese.")
                .build()
            )
        """
        self._extra_system_prompt = text
        return self

    def before_tool(self, hook) -> "CodyBuilder":
        """Register a before-tool hook.

        The hook is called before every tool execution with the signature::

            async def my_hook(tool_name: str, args: dict) -> dict | None:
                ...

        Return the (possibly modified) args dict to proceed, or ``None``
        to reject the call (the model will see a retry message).

        Example::

            async def log_calls(tool_name, args):
                print(f"Calling {tool_name}")
                return args  # proceed unchanged

            client = Cody().before_tool(log_calls).build()
        """
        self._before_tool_hooks.append(hook)
        return self

    def after_tool(self, hook) -> "CodyBuilder":
        """Register an after-tool hook.

        The hook is called after every tool execution with the signature::

            async def my_hook(tool_name: str, args: dict, result: str) -> str:
                ...

        Return the (possibly modified) result string.

        Example::

            async def redact_secrets(tool_name, args, result):
                return result.replace(os.environ["SECRET"], "***")

            client = Cody().after_tool(redact_secrets).build()
        """
        self._after_tool_hooks.append(hook)
        return self

    def session_store(self, store) -> "CodyBuilder":
        """Inject a custom session store (must satisfy SessionStoreProtocol)."""
        self._session_store = store
        return self

    def audit_logger(self, logger) -> "CodyBuilder":
        """Inject a custom audit logger (must satisfy AuditLoggerProtocol)."""
        self._audit_logger = logger
        return self

    def file_history(self, history) -> "CodyBuilder":
        """Inject a custom file history (must satisfy FileHistoryProtocol)."""
        self._file_history = history
        return self

    def memory_store(self, store) -> "CodyBuilder":
        """Inject a custom memory store (must satisfy MemoryStoreProtocol)."""
        self._memory_store = store
        return self

    def stateless(self) -> "CodyBuilder":
        """Enable stateless mode — no persistence (session, audit, file history, memory).

        Uses null storage implementations so the code path stays the same
        but nothing is written to disk. Individual storage can still be
        overridden after calling stateless() (e.g. ``.stateless().audit_logger(real_logger)``).
        """
        self._stateless = True
        return self

    def on(self, event_type: str, handler) -> "CodyBuilder":
        """Register event handler. Implicitly enables events.

        Args:
            event_type: Event type string, e.g. "tool_call", "tool_result".
            handler: Callback function that receives the event.
        """
        self._enable_events = True
        self._event_handlers.append((event_type, handler))
        return self

    def build(self) -> "AsyncCodyClient":
        """Build and return the client instance."""
        cfg = make_config(
            model=self._model,
            workdir=self._workdir,
            api_key=self._api_key,
            base_url=self._base_url,
            enable_thinking=self._enable_thinking,
            thinking_budget=self._thinking_budget,
            permissions=self._permissions,
            allowed_roots=self._allowed_roots,
            strict_read_boundary=self._strict_read_boundary,
            db_path=self._db_path,
            enable_metrics=self._enable_metrics,
            enable_events=self._enable_events,
        )
        if self._interaction is not None:
            cfg.interaction = self._interaction
        if self._circuit_breaker is not None:
            cfg.circuit_breaker = self._circuit_breaker
        cfg.skill_dirs = self._skill_dirs
        cfg.mcp.servers = self._mcp_servers
        cfg.lsp.languages = self._lsp_languages
        # Apply stateless defaults for any storage not explicitly set
        session_store = self._session_store
        audit_logger = self._audit_logger
        file_history = self._file_history
        memory_store = self._memory_store
        if self._stateless:
            from ..core.storage import (
                NullSessionStore, NullAuditLogger,
                NullFileHistory, NullMemoryStore,
            )
            if session_store is UNSET:
                session_store = NullSessionStore()
            if audit_logger is UNSET:
                audit_logger = NullAuditLogger()
            if file_history is UNSET:
                file_history = NullFileHistory()
            if memory_store is UNSET:
                memory_store = NullMemoryStore()

        client = AsyncCodyClient(
            config=cfg,
            auto_start_mcp=self._auto_start_mcp,
            custom_tools=self._custom_tools or None,
            system_prompt=self._system_prompt,
            extra_system_prompt=self._extra_system_prompt,
            before_tool_hooks=self._before_tool_hooks or None,
            after_tool_hooks=self._after_tool_hooks or None,
            session_store=session_store,
            audit_logger=audit_logger,
            file_history=file_history,
            memory_store=memory_store,
        )
        # Apply deferred event handlers
        for event_type_str, handler in self._event_handlers:
            client.on(event_type_str, handler)
        return client


def Cody() -> CodyBuilder:
    """Create a new CodyBuilder instance.

    Usage:
        client = Cody().workdir(".").model("your-model-name").base_url("https://api.example.com/v1").build()
    """
    return CodyBuilder()


# ── Async Client (wraps core directly) ───────────────────────────────────────


class AsyncCodyClient:
    """Async Python SDK for Cody — wraps core engine directly.

    Supports three construction styles:
        # 1. Builder (recommended)
        client = Cody().workdir(".").model("...").build()

        # 2. Direct parameters
        client = AsyncCodyClient(workdir=".", model="...")

        # 3. Config object
        client = AsyncCodyClient(config=cfg)

    Enhanced features (vs. bare core):
        - Event hooks via on() / on_async()
        - Metrics collection
        - Convenience methods (read_file, write_file, etc.)
        - Rich error hierarchy
    """

    def __init__(
        self,
        config: Optional[SDKConfig] = None,
        workdir: Optional[str] = None,
        *,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        db_path: Optional[str] = None,
        enable_metrics: bool = False,
        enable_events: bool = False,
        auto_start_mcp: bool = False,
        custom_tools: list | None = None,
        system_prompt: str | None = None,
        extra_system_prompt: str | None = None,
        before_tool_hooks: list | None = None,
        after_tool_hooks: list | None = None,
        session_store: SessionStoreProtocol | _UnsetType | None = UNSET,
        audit_logger: AuditLoggerProtocol | _UnsetType | None = UNSET,
        file_history: FileHistoryProtocol | _UnsetType | None = UNSET,
        memory_store: MemoryStoreProtocol | _UnsetType | None = UNSET,
    ):
        if config:
            self._config = config
        else:
            self._config = make_config(
                model=model,
                workdir=workdir,
                api_key=api_key,
                base_url=base_url,
                db_path=db_path,
                enable_metrics=enable_metrics,
                enable_events=enable_events,
            )

        self.workdir = Path(self._config.workdir) if self._config.workdir else Path.cwd()
        # Only override model if the user explicitly provided one
        self._model_override = (
            self._config.model.model if self._config.model.model else None
        )
        self._db_path = Path(self._config.db_path) if self._config.db_path else None

        # Core objects (lazy-initialized)
        self._runner = None
        self._session_store = None
        self._core_config = None

        # Custom tools (user-defined async functions)
        self._custom_tools: list = custom_tools or []

        # Custom system prompt overrides
        self._system_prompt: str | None = system_prompt
        self._extra_system_prompt: str | None = extra_system_prompt

        # Step hooks
        self._before_tool_hooks: list = before_tool_hooks or []
        self._after_tool_hooks: list = after_tool_hooks or []

        # Storage layer injection (None = disable, UNSET = use defaults)
        self._injected_session_store = session_store
        self._injected_audit_logger = audit_logger
        self._injected_file_history = file_history
        self._injected_memory_store = memory_store

        # MCP auto-start flag
        self._auto_start_mcp = auto_start_mcp
        self._mcp_started = False

        # LSP auto-start: start configured language servers on first run
        self._lsp_started = False

        # Enhanced features
        self._metrics: Optional[MetricsCollector] = None
        self._events: Optional[EventManager] = None

        if self._config.enable_metrics:
            self._metrics = MetricsCollector()
        if self._config.enable_events:
            self._events = EventManager()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    # ── Core access ──────────────────────────────────────────────────────

    def _get_config(self):
        """Get or create core Config."""
        if self._core_config is None:
            from ..core.config import Config, MCPServerConfig as CoreMCPServerConfig
            self._core_config = Config.load(workdir=self.workdir)
            if self._model_override:
                self._core_config.model = self._model_override
            if self._config.model.enable_thinking:
                self._core_config.enable_thinking = True
                if self._config.model.thinking_budget:
                    self._core_config.thinking_budget = self._config.model.thinking_budget
            if self._config.model.api_key:
                self._core_config.model_api_key = self._config.model.api_key
            if self._config.model.base_url:
                self._core_config.model_base_url = self._config.model.base_url
            # Apply security config from SDK
            if self._config.security.allowed_roots:
                self._core_config.security.allowed_roots = self._config.security.allowed_roots
            if self._config.security.strict_read_boundary:
                self._core_config.security.strict_read_boundary = True
            if self._config.security.blocked_commands:
                self._core_config.security.blocked_commands = self._config.security.blocked_commands
            # Apply custom skill directories from SDK config
            if self._config.skill_dirs:
                existing = set(self._core_config.skills.custom_dirs)
                for d in self._config.skill_dirs:
                    if d not in existing:
                        self._core_config.skills.custom_dirs.append(d)
                        existing.add(d)
            # Apply interaction config from SDK
            if self._config.interaction.enabled:
                from ..core.config import InteractionConfig as CoreIAConfig
                self._core_config.interaction = CoreIAConfig(
                    enabled=self._config.interaction.enabled,
                    timeout=self._config.interaction.timeout,
                )
            # Apply circuit breaker config from SDK
            from ..core.config import CircuitBreakerConfig as CoreCBConfig
            sdk_cb = self._config.circuit_breaker
            # Default SDKCircuitBreakerConfig values for comparison
            defaults = SDKCircuitBreakerConfig()
            if (sdk_cb.enabled != defaults.enabled
                    or sdk_cb.max_tokens != defaults.max_tokens
                    or sdk_cb.max_cost_usd != defaults.max_cost_usd
                    or sdk_cb.max_steps != defaults.max_steps
                    or sdk_cb.loop_detect_turns != defaults.loop_detect_turns
                    or sdk_cb.loop_similarity_threshold != defaults.loop_similarity_threshold
                    or sdk_cb.model_prices):
                cb_dict = sdk_cb.to_dict()
                # Merge model_prices: core defaults ← SDK overrides
                merged_prices = dict(self._core_config.circuit_breaker.model_prices)
                merged_prices.update(cb_dict.pop("model_prices", {}))
                self._core_config.circuit_breaker = CoreCBConfig(
                    **cb_dict, model_prices=merged_prices,
                )
            # Apply MCP servers from SDK config
            if self._config.mcp.enabled and self._config.mcp.servers:
                for s in self._config.mcp.servers:
                    if isinstance(s, SDKMCPServerConfig):
                        d = s.to_dict()
                    else:
                        d = s
                    self._core_config.mcp.servers.append(CoreMCPServerConfig(**d))
        return self._core_config

    def set_config(self, config) -> None:
        """Apply a pre-built core Config, replacing any cached config/runner.

        This is the public API for CLI/TUI to inject a Config with overrides
        (thinking, extra_roots, etc.) without reaching into private attributes.
        """
        self._core_config = config
        self._runner = None  # Force lazy re-creation with updated config

    def get_runner(self):
        """Get or create the underlying AgentRunner.

        Power-user API for callers that need raw streaming events
        (ToolCallEvent, ThinkingEvent, etc.) or direct MCP control.
        """
        if self._runner is None:
            from ..core.runner import AgentRunner
            self._runner = AgentRunner(
                config=self._get_config(),
                workdir=self.workdir,
                custom_tools=self._custom_tools or None,
                system_prompt=self._system_prompt,
                extra_system_prompt=self._extra_system_prompt,
                before_tool_hooks=self._before_tool_hooks or None,
                after_tool_hooks=self._after_tool_hooks or None,
                audit_logger=self._injected_audit_logger,
                file_history=self._injected_file_history,
                memory_store=self._injected_memory_store,
            )
        return self._runner

    def get_session_store(self):
        """Get or create the underlying SessionStore.

        Power-user API for callers that need synchronous session access
        (e.g. TUI on_mount) or direct store operations.

        If a custom session store was injected via the builder, it is returned
        directly (must satisfy ``SessionStoreProtocol``).
        """
        if self._session_store is None:
            injected = self._injected_session_store
            if injected is not UNSET and injected is not None:
                # Custom store injected via builder
                self._session_store = injected
            else:
                # UNSET or None — use default SQLite store
                from ..core.session import SessionStore
                self._session_store = SessionStore(db_path=self._db_path)
        return self._session_store

    async def start_mcp(self) -> None:
        """Start MCP servers if configured. Called automatically on first run() when auto_start_mcp=True."""
        runner = self.get_runner()
        if not self._mcp_started:
            await runner.start_mcp()
            self._mcp_started = True

    async def start_lsp(self) -> None:
        """Start LSP servers for configured languages.

        Called automatically on first run()/stream() when lsp.enabled is True
        and lsp.languages is non-empty.  Silently skips languages whose
        server binary is not installed.
        """
        if self._lsp_started:
            return
        self._lsp_started = True
        languages = self._config.lsp.languages if self._config.lsp.enabled else []
        if not languages:
            return
        runner = self.get_runner()
        for lang in languages:
            try:
                await runner.start_lsp(lang)
            except Exception:
                # LSP server not installed — skip silently
                pass

    async def add_mcp_server(
        self,
        name: str,
        *,
        transport: Literal["stdio", "http"] = "stdio",
        command: str = "",
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
        url: str = "",
        headers: dict[str, str] | None = None,
    ) -> None:
        """Dynamically add and start an MCP server at runtime.

        Args:
            name: Server name (used as prefix in tool calls, e.g. "feishu/fetch-doc").
            transport: "stdio" or "http".
            command: Command to run (stdio only).
            args: Command arguments (stdio only).
            env: Environment variables (stdio only).
            url: HTTP endpoint URL (http only).
            headers: HTTP headers (http only).
        """
        from ..core.config import MCPServerConfig as CoreMCPServerConfig

        server_config = CoreMCPServerConfig(
            name=name,
            transport=transport,
            command=command,
            args=args or [],
            env=env or {},
            url=url,
            headers=headers or {},
        )

        runner = self.get_runner()
        # Add to core config so it persists
        runner.config.mcp.servers.append(server_config)
        # Start this single server immediately
        if runner._mcp_client:
            await runner._mcp_client.start_server(server_config)
        else:
            # MCP client not yet created, start full MCP
            await runner.start_mcp()
            self._mcp_started = True

    async def close(self):
        """Clean up resources."""
        if self._runner:
            await self._runner.stop_mcp()
            await self._runner.stop_lsp()
            self._runner = None
        if self._events:
            await self._events.dispatch_async(SessionEvent(
                event_type=EventType.SESSION_CLOSE,
            ))

    # ── Health ────────────────────────────────────────────────────────────

    async def health(self) -> dict:
        """Return SDK health info."""
        from .. import __version__
        return {"status": "ok", "version": __version__}

    # ── Run ───────────────────────────────────────────────────────────────

    @overload
    async def run(
        self, prompt, *, session_id: Optional[str] = None, stream: Literal[False] = False,
        include_tools: list[str] | None = None, exclude_tools: list[str] | None = None,
        cancel_event: Optional[asyncio.Event] = None,
    ) -> RunResult: ...

    @overload
    async def run(
        self, prompt, *, session_id: Optional[str] = None, stream: Literal[True],
        include_tools: list[str] | None = None, exclude_tools: list[str] | None = None,
        cancel_event: Optional[asyncio.Event] = None,
    ) -> AsyncIterator[StreamChunk]: ...

    async def run(
        self,
        prompt,
        *,
        session_id: Optional[str] = None,
        stream: bool = False,
        include_tools: list[str] | None = None,
        exclude_tools: list[str] | None = None,
        cancel_event: Optional[asyncio.Event] = None,
    ) -> RunResult | AsyncIterator[StreamChunk]:
        """Run agent with prompt.

        Args:
            prompt: Task description (str or Prompt).
            session_id: Optional session ID for multi-turn.
            stream: If True, return async iterator of StreamChunk.
            include_tools: If set, only these tools are available for this run.
            exclude_tools: If set, these tools are excluded for this run.
            cancel_event: If set and triggered, cancels the run.
                Non-streaming: returns ``RunResult(output="(cancelled)")``.
                Streaming: yields a ``cancelled`` chunk.

        Returns:
            RunResult if stream=False, else AsyncIterator[StreamChunk].
        """
        # Auto-start MCP servers on first run (if enabled)
        if self._auto_start_mcp and not self._mcp_started:
            await self.start_mcp()

        # Auto-start LSP servers for configured languages
        if not self._lsp_started:
            await self.start_lsp()

        # Fire event
        if self._events:
            await self._events.dispatch_async(RunEvent(
                event_type=EventType.RUN_START,
                prompt=str(prompt),
                session_id=session_id,
            ))

        try:
            if stream:
                return self._stream_run(
                    prompt, session_id,
                    include_tools=include_tools, exclude_tools=exclude_tools,
                    cancel_event=cancel_event,
                )

            # Start metrics (non-streaming only; streaming handles its own)
            if self._metrics:
                self._metrics.start_run(
                    str(prompt), session_id, self._config.model.enable_thinking
                )

            # Emit MODEL_REQUEST before calling the model
            if self._events:
                await self._events.dispatch_async(ModelEvent(
                    event_type=EventType.MODEL_REQUEST,
                    model=self._config.model.model or "",
                ))

            runner = self.get_runner()
            # Always use session to enable multi-turn by default
            store = self.get_session_store()
            result, sid = await runner.run_with_session(
                prompt, store, session_id,
                include_tools=include_tools, exclude_tools=exclude_tools,
                cancel_event=cancel_event,
            )

            run_result = RunResult(
                output=result.output,
                session_id=sid,
                usage=_usage_from_result(result),
                thinking=result.thinking,
                metadata=result.metadata,
            )

            # Emit MODEL_RESPONSE with usage
            if self._events:
                await self._events.dispatch_async(ModelEvent(
                    event_type=EventType.MODEL_RESPONSE,
                    model=self._config.model.model or "",
                    input_tokens=run_result.usage.input_tokens,
                    output_tokens=run_result.usage.output_tokens,
                ))

            # Record metrics
            if self._metrics:
                self._metrics.end_run(
                    result.output,
                    TokenUsage(
                        input_tokens=run_result.usage.input_tokens,
                        output_tokens=run_result.usage.output_tokens,
                        total_tokens=run_result.usage.total_tokens,
                    ),
                )

            # Fire tool events from traces
            if self._events and result.tool_traces:
                for trace in result.tool_traces:
                    await self._events.dispatch_async(ToolEvent(
                        event_type=EventType.TOOL_CALL,
                        tool_name=trace.tool_name,
                        args=trace.args,
                    ))
                    await self._events.dispatch_async(ToolEvent(
                        event_type=EventType.TOOL_RESULT,
                        tool_name=trace.tool_name,
                        result=trace.result,
                    ))

            # Fire run end event
            if self._events:
                await self._events.dispatch_async(RunEvent(
                    event_type=EventType.RUN_END,
                    prompt=str(prompt),
                    session_id=session_id,
                    result=result.output,
                ))

            return run_result

        except Exception as e:
            # Ensure metrics run is closed on exception
            if self._metrics and self._metrics._current_run is not None:
                self._metrics.end_run("", TokenUsage(0, 0, 0))
            if self._events:
                await self._events.dispatch_async(ModelEvent(
                    event_type=EventType.MODEL_ERROR,
                    model=self._config.model.model or "",
                    error=str(e),
                ))
                await self._events.dispatch_async(RunEvent(
                    event_type=EventType.RUN_ERROR,
                    prompt=str(prompt),
                    session_id=session_id,
                    error=str(e),
                ))
            raise

    async def stream(
        self,
        prompt,
        *,
        session_id: Optional[str] = None,
        cancel_event: Optional[asyncio.Event] = None,
        include_tools: list[str] | None = None,
        exclude_tools: list[str] | None = None,
    ) -> AsyncIterator[StreamChunk]:
        """Stream agent response. Yields StreamChunk objects.

        SDK EventType events (STREAM_START/END, THINKING_START/END, TOOL_CALL,
        TOOL_RESULT, etc.) are dispatched automatically when an EventManager
        is configured.
        """
        # Auto-start MCP servers on first stream (if enabled)
        if self._auto_start_mcp and not self._mcp_started:
            await self.start_mcp()

        # Auto-start LSP servers for configured languages
        if not self._lsp_started:
            await self.start_lsp()

        runner = self.get_runner()
        store = self.get_session_store()

        in_thinking = False
        stream_started = False

        async for event, sid in runner.run_stream_with_session(
            prompt, store, session_id, cancel_event=cancel_event,
            include_tools=include_tools, exclude_tools=exclude_tools,
        ):
            chunk = _event_to_chunk(event, sid)

            # ── SDK event dispatch ──
            if self._events:
                # Emit STREAM_START on first non-session_start chunk
                if not stream_started and chunk.type != "session_start":
                    stream_started = True
                    await self._events.dispatch_async(SDKStreamEvent(
                        event_type=EventType.STREAM_START,
                        chunk_type="stream_start",
                    ))

                if chunk.type == "thinking":
                    if not in_thinking:
                        in_thinking = True
                        await self._events.dispatch_async(SDKThinkingEvent(
                            event_type=EventType.THINKING_START,
                            content="",
                            is_start=True,
                        ))
                    await self._events.dispatch_async(SDKThinkingEvent(
                        event_type=EventType.THINKING_CHUNK,
                        content=chunk.content,
                    ))
                else:
                    # End thinking block when a non-thinking chunk arrives
                    if in_thinking:
                        in_thinking = False
                        await self._events.dispatch_async(SDKThinkingEvent(
                            event_type=EventType.THINKING_END,
                            content="",
                            is_end=True,
                        ))

                    if chunk.type == "tool_call":
                        await self._events.dispatch_async(ToolEvent(
                            event_type=EventType.TOOL_CALL,
                            tool_name=chunk.tool_name or chunk.content,
                            args=chunk.args or {},
                        ))
                    elif chunk.type == "tool_result":
                        await self._events.dispatch_async(ToolEvent(
                            event_type=EventType.TOOL_RESULT,
                            tool_name=chunk.tool_name or "",
                            result=chunk.content,
                        ))
                    elif chunk.type == "text_delta":
                        await self._events.dispatch_async(SDKStreamEvent(
                            event_type=EventType.STREAM_CHUNK,
                            chunk_type=chunk.type,
                            content=chunk.content,
                        ))
                    elif chunk.type == "compact":
                        await self._events.dispatch_async(SDKContextCompactEvent(
                            event_type=EventType.CONTEXT_COMPACT,
                            original_messages=chunk.original_messages,
                            compacted_messages=chunk.compacted_messages,
                            tokens_saved=chunk.estimated_tokens_saved,
                        ))
                    elif chunk.type == "done":
                        usage = chunk.usage
                        await self._events.dispatch_async(ModelEvent(
                            event_type=EventType.MODEL_RESPONSE,
                            model=self._config.model.model or "",
                            input_tokens=usage.input_tokens if usage else 0,
                            output_tokens=usage.output_tokens if usage else 0,
                        ))

                # Emit STREAM_END on done/cancelled/circuit_breaker
                if chunk.type in ("done", "cancelled", "circuit_breaker"):
                    if in_thinking:
                        in_thinking = False
                        await self._events.dispatch_async(SDKThinkingEvent(
                            event_type=EventType.THINKING_END,
                            content="",
                            is_end=True,
                        ))
                    await self._events.dispatch_async(SDKStreamEvent(
                        event_type=EventType.STREAM_END,
                        chunk_type="stream_end",
                    ))

            yield chunk

    # Alias for stream() — matches the name used in demos/docs
    run_stream = stream

    async def _stream_run(
        self, prompt, session_id: Optional[str] = None,
        include_tools: list[str] | None = None,
        exclude_tools: list[str] | None = None,
        cancel_event: Optional[asyncio.Event] = None,
    ):
        """Internal streaming run (called when run(stream=True)).

        Wraps stream() with metrics tracking so that run(stream=True)
        properly closes the metrics run when the stream completes.
        """
        if self._metrics:
            self._metrics.start_run(
                str(prompt), session_id, self._config.model.enable_thinking
            )
        try:
            last_chunk = None
            async for chunk in self.stream(
                prompt, session_id=session_id,
                include_tools=include_tools, exclude_tools=exclude_tools,
                cancel_event=cancel_event,
            ):
                last_chunk = chunk
                yield chunk
        finally:
            if self._metrics and self._metrics._current_run is not None:
                # Extract usage from the final done chunk if available
                usage = TokenUsage(0, 0, 0)
                if last_chunk and hasattr(last_chunk, 'usage') and last_chunk.usage:
                    usage = TokenUsage(
                        input_tokens=last_chunk.usage.input_tokens,
                        output_tokens=last_chunk.usage.output_tokens,
                        total_tokens=last_chunk.usage.total_tokens,
                    )
                output = ""
                if last_chunk and hasattr(last_chunk, 'content'):
                    output = last_chunk.content or ""
                self._metrics.end_run(output, usage)

    # ── Tool ──────────────────────────────────────────────────────────────

    async def tool(
        self,
        tool_name: str,
        params: Optional[dict] = None,
        *,
        workdir: Optional[str] = None,
    ) -> ToolResult:
        """Call a tool directly."""
        from ..core import tools
        from ..core.deps import CodyDeps, ToolContext
        from ..core.skill_manager import SkillManager

        tool_func = getattr(tools, tool_name, None)
        if not tool_func:
            raise CodyNotFoundError(
                f"Tool not found: {tool_name}",
                code="TOOL_NOT_FOUND",
            )

        if self._events:
            await self._events.dispatch_async(ToolEvent(
                event_type=EventType.TOOL_CALL,
                tool_name=tool_name,
                args=params or {},
            ))

        effective_workdir = Path(workdir) if workdir else self.workdir
        cfg = self._get_config()
        sm = SkillManager(config=cfg, workdir=effective_workdir)
        # Build deps directly from client state — no need to create an
        # AgentRunner (which requires model_base_url) just to call a tool.
        fh = self._injected_file_history if not isinstance(self._injected_file_history, _UnsetType) else None
        al = self._injected_audit_logger if not isinstance(self._injected_audit_logger, _UnsetType) else None
        deps = CodyDeps(
            config=cfg,
            workdir=effective_workdir,
            skill_manager=sm,
            allowed_roots=[effective_workdir],
            strict_read_boundary=cfg.security.strict_read_boundary,
            file_history=fh,
            audit_logger=al,
        )

        start_time = time.time()
        try:
            result_str = await tool_func(ToolContext(deps), **(params or {}))
            duration = time.time() - start_time

            if self._metrics:
                self._metrics.record_tool_call(tool_name, duration, success=True)
            if self._events:
                await self._events.dispatch_async(ToolEvent(
                    event_type=EventType.TOOL_RESULT,
                    tool_name=tool_name,
                    args=params or {},
                    result=result_str[:500] if result_str else "",
                    duration=duration,
                ))
            return ToolResult(result=result_str)

        except Exception as e:
            duration = time.time() - start_time
            if self._metrics:
                self._metrics.record_tool_call(
                    tool_name, duration, success=False, error=str(e)
                )
            if self._events:
                await self._events.dispatch_async(ToolEvent(
                    event_type=EventType.TOOL_ERROR,
                    tool_name=tool_name,
                    args=params or {},
                    error=str(e),
                    duration=duration,
                ))
            raise CodyToolError(str(e), tool_name=tool_name) from e

    # ── Sessions ──────────────────────────────────────────────────────────

    async def create_session(
        self,
        title: str = "New session",
        model: str = "",
        workdir: str = "",
    ) -> SessionInfo:
        """Create a new session."""
        store = self.get_session_store()
        session = store.create_session(
            title=title,
            model=model,
            workdir=workdir or str(self.workdir),
        )
        info = SessionInfo(
            id=session.id,
            title=session.title,
            model=session.model,
            workdir=session.workdir,
            message_count=len(session.messages),
            created_at=session.created_at,
            updated_at=session.updated_at,
        )

        if self._events:
            from .events import SessionEvent
            await self._events.dispatch_async(SessionEvent(
                event_type=EventType.SESSION_CREATE,
                session_id=session.id,
                title=title,
            ))

        return info

    async def list_sessions(self, limit: int = 20) -> list[SessionInfo]:
        """List recent sessions."""
        store = self.get_session_store()
        sessions = store.list_sessions(limit=limit)
        return [
            SessionInfo(
                id=s.id,
                title=s.title,
                model=s.model,
                workdir=s.workdir,
                message_count=s.message_count if s.message_count is not None else len(s.messages),
                created_at=s.created_at,
                updated_at=s.updated_at,
            )
            for s in sessions
        ]

    async def get_session(self, session_id: str) -> SessionDetail:
        """Get session with messages."""
        store = self.get_session_store()
        session = store.get_session(session_id)
        if not session:
            raise CodyNotFoundError(
                f"Session not found: {session_id}",
                code="SESSION_NOT_FOUND",
            )
        return SessionDetail(
            id=session.id,
            title=session.title,
            model=session.model,
            workdir=session.workdir,
            message_count=len(session.messages),
            created_at=session.created_at,
            updated_at=session.updated_at,
            messages=[
                {"role": m.role, "content": m.content, "timestamp": m.timestamp}
                for m in session.messages
            ],
        )

    async def delete_session(self, session_id: str) -> None:
        """Delete a session."""
        store = self.get_session_store()
        deleted = store.delete_session(session_id)
        if not deleted:
            raise CodyNotFoundError(
                f"Session not found: {session_id}",
                code="SESSION_NOT_FOUND",
            )

    async def get_latest_session(
        self,
        workdir: str | None = None,
    ) -> SessionInfo | None:
        """Get the most recent session, optionally filtered by workdir."""
        store = self.get_session_store()
        session = store.get_latest_session(workdir=workdir)
        if not session:
            return None
        return SessionInfo(
            id=session.id,
            title=session.title,
            model=session.model,
            workdir=session.workdir,
            message_count=len(session.messages),
            created_at=session.created_at,
            updated_at=session.updated_at,
        )

    def get_message_count(self, session_id: str) -> int:
        """Get message count for a session."""
        store = self.get_session_store()
        return store.get_message_count(session_id)

    def add_message(self, session_id: str, role: str, content: str) -> None:
        """Add a message to a session."""
        store = self.get_session_store()
        store.add_message(session_id, role, content)

    def update_title(self, session_id: str, title: str) -> None:
        """Update session title."""
        store = self.get_session_store()
        store.update_title(session_id, title)

    @staticmethod
    def messages_to_history(messages) -> list:
        """Convert stored session messages to pydantic-ai message format."""
        from ..core.runner import AgentRunner
        return AgentRunner.messages_to_history(messages)

    # ── Skills ────────────────────────────────────────────────────────────

    async def list_skills(self) -> list[dict]:
        """List available skills."""
        from ..core.skill_manager import SkillManager
        sm = SkillManager(config=self._get_config(), workdir=self.workdir)
        return [
            {
                "name": s.name,
                "description": s.description,
                "source": s.source,
                "enabled": s.enabled,
            }
            for s in sm.list_skills()
        ]

    async def get_skill(self, skill_name: str) -> dict:
        """Get skill details including full documentation."""
        from ..core.skill_manager import SkillManager
        sm = SkillManager(config=self._get_config(), workdir=self.workdir)
        skill = sm.get_skill(skill_name)
        if not skill:
            raise CodyNotFoundError(
                f"Skill not found: {skill_name}",
                code="SKILL_NOT_FOUND",
            )
        return {
            "name": skill.name,
            "description": skill.description,
            "source": skill.source,
            "enabled": skill.enabled,
            "documentation": skill.documentation,
        }

    async def enable_skill(self, skill_name: str) -> None:
        """Enable a skill and persist the change to config."""
        from ..core.skill_manager import SkillManager
        from ..shared import resolve_config_path
        cfg = self._get_config()
        sm = SkillManager(config=cfg, workdir=self.workdir)
        skill = sm.get_skill(skill_name)
        if not skill:
            raise CodyNotFoundError(
                f"Skill not found: {skill_name}",
                code="SKILL_NOT_FOUND",
            )
        sm.enable_skill(skill_name)
        cfg.save(resolve_config_path())

    async def disable_skill(self, skill_name: str) -> None:
        """Disable a skill and persist the change to config."""
        from ..core.skill_manager import SkillManager
        from ..shared import resolve_config_path
        cfg = self._get_config()
        sm = SkillManager(config=cfg, workdir=self.workdir)
        skill = sm.get_skill(skill_name)
        if not skill:
            raise CodyNotFoundError(
                f"Skill not found: {skill_name}",
                code="SKILL_NOT_FOUND",
            )
        sm.disable_skill(skill_name)
        cfg.save(resolve_config_path())

    # ── Convenience Methods ──────────────────────────────────────────────

    async def read_file(self, path: str) -> str:
        """Read a file."""
        result = await self.tool("read_file", {"path": path})
        return result.result

    async def write_file(self, path: str, content: str) -> str:
        """Write a file."""
        result = await self.tool("write_file", {"path": path, "content": content})
        return result.result

    async def edit_file(self, path: str, old_text: str, new_text: str) -> str:
        """Edit a file."""
        result = await self.tool("edit_file", {
            "path": path, "old_text": old_text, "new_text": new_text,
        })
        return result.result

    async def list_directory(self, path: str = ".") -> str:
        """List directory contents."""
        result = await self.tool("list_directory", {"path": path})
        return result.result

    async def grep(self, pattern: str, include: str = "*") -> str:
        """Search for pattern in files."""
        result = await self.tool("grep", {"pattern": pattern, "include": include})
        return result.result

    async def glob(self, pattern: str) -> str:
        """Find files by glob pattern."""
        result = await self.tool("glob", {"pattern": pattern})
        return result.result

    async def exec_command(self, command: str) -> str:
        """Execute shell command."""
        result = await self.tool("exec_command", {"command": command})
        return result.result

    async def search_files(self, query: str) -> str:
        """Search for files by name (fuzzy)."""
        result = await self.tool("search_files", {"query": query})
        return result.result

    # ── MCP Methods ─────────────────────────────────────────────────────

    async def mcp_list_tools(self) -> list[dict]:
        """List all tools from connected MCP servers."""
        runner = self.get_runner()
        if not runner._mcp_client:
            return []
        return [
            {
                "name": f"{t.server_name}/{t.name}",
                "description": t.description,
                "input_schema": t.input_schema,
                "server": t.server_name,
            }
            for t in runner._mcp_client.list_tools()
        ]

    async def mcp_call(
        self,
        tool_name: str,
        arguments: dict | None = None,
    ) -> str:
        """Call an MCP tool by 'server_name/tool_name'."""
        runner = self.get_runner()
        if not runner._mcp_client:
            raise CodyConfigError("No MCP servers configured or started")
        result = await runner._mcp_client.call_tool(tool_name, arguments)
        return result

    # ── LSP Methods ──────────────────────────────────────────────────────

    async def lsp_diagnostics(self, file_path: str) -> str:
        """Get LSP diagnostics for a file."""
        result = await self.tool("lsp_diagnostics", {"file_path": file_path})
        return result.result

    async def lsp_definition(self, file_path: str, line: int, column: int) -> str:
        """Go to definition."""
        result = await self.tool("lsp_definition", {
            "file_path": file_path, "line": line, "character": column,
        })
        return result.result

    async def lsp_references(self, file_path: str, line: int, column: int) -> str:
        """Find references."""
        result = await self.tool("lsp_references", {
            "file_path": file_path, "line": line, "character": column,
        })
        return result.result

    async def lsp_hover(self, file_path: str, line: int, column: int) -> str:
        """Get hover info."""
        result = await self.tool("lsp_hover", {
            "file_path": file_path, "line": line, "character": column,
        })
        return result.result

    # ── Interaction Methods ──────────────────────────────────────────────

    async def submit_interaction(
        self,
        request_id: str,
        action: Literal["approve", "reject", "revise", "answer"] = "answer",
        content: str = "",
    ) -> None:
        """Submit a human response to a pending interaction request.

        Args:
            request_id: The ID of the InteractionRequest to respond to.
            action: One of "approve", "reject", "revise", "answer".
            content: Response content (e.g., the user's answer or revision).
        """
        from ..core.interaction import InteractionResponse
        runner = self.get_runner()
        response = InteractionResponse(
            request_id=request_id, action=action, content=content,
        )
        await runner.submit_interaction(response)

    async def inject_user_input(self, message: str) -> None:
        """Send a proactive message to the running agent without waiting for it to ask.

        The message is queued and injected at the next node boundary
        (after current tool execution), so the LLM sees it on the next turn.

        Args:
            message: The text to inject into the conversation.
        """
        runner = self.get_runner()
        await runner.inject_user_input(message)

    # ── Memory Methods ────────────────────────────────────────────────────

    async def add_memory(
        self,
        category: str,
        content: str,
        *,
        source_task_id: str = "",
        source_task_title: str = "",
        confidence: float = 1.0,
        tags: list[str] | None = None,
    ) -> None:
        """Add a memory entry to the project memory store.

        Args:
            category: One of "conventions", "patterns", "issues", "decisions".
            content: The memory content.
            source_task_id: Optional task ID that produced this memory.
            source_task_title: Optional task title.
            confidence: Confidence score (0.0-1.0).
            tags: Optional tags for filtering.
        """
        from ..core.memory import MemoryEntry
        runner = self.get_runner()
        if not runner.memory_store:
            return
        entry = MemoryEntry(
            content=content,
            source_task_id=source_task_id,
            source_task_title=source_task_title,
            confidence=confidence,
            tags=tags or [],
        )
        await runner.memory_store.add_entries(category, [entry])

    async def get_memory(self) -> dict[str, list[dict]]:
        """Get all project memory entries grouped by category."""
        runner = self.get_runner()
        if not runner.memory_store:
            return {}
        all_entries = runner.memory_store.get_all_entries()
        return {
            cat: [
                {
                    "id": e.id,
                    "content": e.content,
                    "confidence": e.confidence,
                    "tags": e.tags,
                    "created_at": e.created_at,
                }
                for e in entries
            ]
            for cat, entries in all_entries.items()
        }

    async def clear_memory(self) -> None:
        """Clear all project memory."""
        runner = self.get_runner()
        if runner.memory_store:
            runner.memory_store.clear()

    # ── Event Methods ────────────────────────────────────────────────────

    def on(self, event_type, handler):
        """Register event handler.

        Args:
            event_type: EventType enum or string (e.g. "tool_call").
            handler: Callback function.
        """
        if not self._events:
            raise CodyConfigError("Events not enabled. Use enable_events() in config.")
        if isinstance(event_type, str):
            event_type = EventType(event_type)
        self._events.register(event_type, handler)

    def on_async(self, event_type, handler):
        """Register async event handler.

        Args:
            event_type: EventType enum or string (e.g. "tool_call").
            handler: Callback function.
        """
        if not self._events:
            raise CodyConfigError("Events not enabled. Use enable_events() in config.")
        if isinstance(event_type, str):
            event_type = EventType(event_type)
        self._events.register_async(event_type, handler)

    # ── Metrics Methods ──────────────────────────────────────────────────

    def get_metrics(self) -> Optional[dict]:
        """Get metrics summary."""
        if not self._metrics:
            return None
        return self._metrics.get_summary()

    def get_metrics_collector(self) -> Optional[MetricsCollector]:
        """Get metrics collector instance."""
        return self._metrics


# ── Sync Client ──────────────────────────────────────────────────────────────


def _run_async(coro):
    """Run an async coroutine from sync context."""
    try:
        asyncio.get_running_loop()
        import concurrent.futures
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as pool:
            return pool.submit(asyncio.run, coro).result()
    except RuntimeError:
        return asyncio.run(coro)


class CodyClient:
    """Synchronous Python SDK for Cody — wraps AsyncCodyClient.

    Usage:
        with CodyClient(workdir=".") as client:
            result = client.run("task")
    """

    def __init__(self, **kwargs):
        self._async = AsyncCodyClient(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def start_mcp(self) -> None:
        """Start MCP servers if configured."""
        _run_async(self._async.start_mcp())

    def close(self):
        _run_async(self._async.close())

    def health(self) -> dict:
        return _run_async(self._async.health())

    def run(self, prompt, *, session_id: Optional[str] = None):
        return _run_async(self._async.run(prompt, session_id=session_id))

    def stream(self, prompt, *, session_id: Optional[str] = None):
        """Collect all stream chunks (sync version returns list)."""
        async def _collect():
            chunks = []
            async for chunk in self._async.stream(prompt, session_id=session_id):
                chunks.append(chunk)
            return chunks
        return _run_async(_collect())

    # Alias
    run_stream = stream

    def tool(self, tool_name: str, params: Optional[dict] = None, **kwargs):
        return _run_async(self._async.tool(tool_name, params, **kwargs))

    def create_session(self, title: str = "New session", **kwargs):
        return _run_async(self._async.create_session(title=title, **kwargs))

    def list_sessions(self, limit: int = 20):
        return _run_async(self._async.list_sessions(limit=limit))

    def get_session(self, session_id: str):
        return _run_async(self._async.get_session(session_id))

    def delete_session(self, session_id: str):
        return _run_async(self._async.delete_session(session_id))

    def list_skills(self):
        return _run_async(self._async.list_skills())

    def get_skill(self, skill_name: str):
        return _run_async(self._async.get_skill(skill_name))

    def get_latest_session(self, workdir: str | None = None):
        return _run_async(self._async.get_latest_session(workdir=workdir))

    def enable_skill(self, skill_name: str):
        return _run_async(self._async.enable_skill(skill_name))

    def disable_skill(self, skill_name: str):
        return _run_async(self._async.disable_skill(skill_name))

    def get_message_count(self, session_id: str) -> int:
        return self._async.get_message_count(session_id)

    def add_message(self, session_id: str, role: str, content: str) -> None:
        self._async.add_message(session_id, role, content)

    def update_title(self, session_id: str, title: str) -> None:
        self._async.update_title(session_id, title)

    @staticmethod
    def messages_to_history(messages) -> list:
        return AsyncCodyClient.messages_to_history(messages)

    def mcp_list_tools(self) -> list[dict]:
        return _run_async(self._async.mcp_list_tools())

    def mcp_call(self, tool_name: str, arguments: dict | None = None) -> str:
        return _run_async(self._async.mcp_call(tool_name, arguments))

    def read_file(self, path: str) -> str:
        return _run_async(self._async.read_file(path))

    def write_file(self, path: str, content: str) -> str:
        return _run_async(self._async.write_file(path, content))

    def get_metrics(self) -> Optional[dict]:
        return self._async.get_metrics()

    def submit_interaction(self, request_id: str, action: str = "answer", content: str = "") -> None:
        _run_async(self._async.submit_interaction(request_id, action, content))

    def inject_user_input(self, message: str) -> None:
        _run_async(self._async.inject_user_input(message))

    def add_memory(self, category: str, content: str, **kwargs) -> None:
        _run_async(self._async.add_memory(category, content, **kwargs))

    def get_memory(self) -> dict:
        return _run_async(self._async.get_memory())

    def clear_memory(self) -> None:
        _run_async(self._async.clear_memory())



"""Tests for rate limiting"""

import time

import pytest

from cody.core.rate_limiter import RateLimiter, RateLimitResult


# ── constructor ──────────────────────────────────────────────────────────────


def test_constructor_defaults():
    limiter = RateLimiter()
    assert limiter.max_requests == 60
    assert limiter.window_seconds == 60.0


def test_constructor_custom():
    limiter = RateLimiter(max_requests=10, window_seconds=30.0)
    assert limiter.max_requests == 10
    assert limiter.window_seconds == 30.0


def test_constructor_invalid_max_requests():
    with pytest.raises(ValueError, match="max_requests"):
        RateLimiter(max_requests=0)


def test_constructor_invalid_window():
    with pytest.raises(ValueError, match="window_seconds"):
        RateLimiter(window_seconds=0)


# ── check (read-only) ───────────────────────────────────────────────────────


def test_check_allowed():
    limiter = RateLimiter(max_requests=5, window_seconds=60.0)
    result = limiter.check("client1")
    assert result.allowed is True
    assert result.remaining == 5
    assert result.limit == 5


def test_check_does_not_consume():
    limiter = RateLimiter(max_requests=2, window_seconds=60.0)
    limiter.check("client1")
    limiter.check("client1")
    limiter.check("client1")
    # check doesn't consume, so hit should still be allowed
    result = limiter.hit("client1")
    assert result.allowed is True


# ── hit ──────────────────────────────────────────────────────────────────────


def test_hit_allowed():
    limiter = RateLimiter(max_requests=3, window_seconds=60.0)
    result = limiter.hit("client1")
    assert result.allowed is True
    assert result.remaining == 2
    assert result.limit == 3


def test_hit_exhausts_limit():
    limiter = RateLimiter(max_requests=3, window_seconds=60.0)

    for _ in range(3):
        result = limiter.hit("client1")
        assert result.allowed is True

    # Fourth request should be denied
    result = limiter.hit("client1")
    assert result.allowed is False
    assert result.remaining == 0
    assert result.retry_after is not None
    assert result.retry_after > 0


def test_hit_different_keys_independent():
    limiter = RateLimiter(max_requests=2, window_seconds=60.0)

    limiter.hit("client1")
    limiter.hit("client1")

    # client2 should be unaffected
    result = limiter.hit("client2")
    assert result.allowed is True
    assert result.remaining == 1


def test_hit_default_key():
    limiter = RateLimiter(max_requests=2, window_seconds=60.0)
    result = limiter.hit()
    assert result.allowed is True


# ── window expiry ────────────────────────────────────────────────────────────


def test_window_expiry():
    limiter = RateLimiter(max_requests=2, window_seconds=0.1)

    limiter.hit("client1")
    limiter.hit("client1")
    result = limiter.hit("client1")
    assert result.allowed is False

    # Wait for window to expire
    time.sleep(0.15)

    result = limiter.hit("client1")
    assert result.allowed is True


# ── reset ────────────────────────────────────────────────────────────────────


def test_reset_specific_key():
    limiter = RateLimiter(max_requests=2, window_seconds=60.0)

    limiter.hit("client1")
    limiter.hit("client1")
    limiter.hit("client2")

    limiter.reset("client1")

    # client1 should be reset
    result = limiter.hit("client1")
    assert result.allowed is True
    assert result.remaining == 1

    # client2 should still have 1 hit
    result = limiter.check("client2")
    assert result.remaining == 1


def test_reset_all():
    limiter = RateLimiter(max_requests=2, window_seconds=60.0)

    limiter.hit("client1")
    limiter.hit("client2")

    limiter.reset()

    result = limiter.hit("client1")
    assert result.allowed is True
    assert result.remaining == 1


# ── RateLimitResult ──────────────────────────────────────────────────────────


def test_rate_limit_result_defaults():
    result = RateLimitResult(allowed=True, remaining=5, limit=10)
    assert result.retry_after is None


def test_rate_limit_result_with_retry():
    result = RateLimitResult(allowed=False, remaining=0, limit=10, retry_after=5.5)
    assert result.retry_after == 5.5


# ── remaining count accuracy ─────────────────────────────────────────────────


def test_remaining_decreases():
    limiter = RateLimiter(max_requests=5, window_seconds=60.0)

    r1 = limiter.hit("key")
    assert r1.remaining == 4

    r2 = limiter.hit("key")
    assert r2.remaining == 3

    r3 = limiter.hit("key")
    assert r3.remaining == 2



"""Cody SDK - Metrics collection.

Provides metrics for monitoring SDK performance and usage.
"""

import time
from dataclasses import dataclass, field
from typing import Optional
from collections import defaultdict


@dataclass
class TokenUsage:
    """Token usage statistics."""
    
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0
    
    def add(self, other: "TokenUsage") -> "TokenUsage":
        """Add another TokenUsage to this one."""
        return TokenUsage(
            input_tokens=self.input_tokens + other.input_tokens,
            output_tokens=self.output_tokens + other.output_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )


@dataclass
class ToolMetrics:
    """Tool execution metrics."""
    
    tool_name: str
    call_count: int = 0
    success_count: int = 0
    error_count: int = 0
    total_duration: float = 0.0
    
    @property
    def avg_duration(self) -> float:
        """Average duration per call."""
        if self.call_count == 0:
            return 0.0
        return self.total_duration / self.call_count
    
    @property
    def success_rate(self) -> float:
        """Success rate (0.0 - 1.0)."""
        if self.call_count == 0:
            return 0.0
        return self.success_count / self.call_count


@dataclass
class RunMetrics:
    """Single run metrics."""
    
    prompt_length: int = 0
    output_length: int = 0
    duration: float = 0.0
    token_usage: TokenUsage = field(default_factory=TokenUsage)
    tool_calls: int = 0
    thinking_enabled: bool = False
    session_id: Optional[str] = None


@dataclass
class SessionMetrics:
    """Session-level metrics."""
    
    session_id: str
    run_count: int = 0
    total_tokens: TokenUsage = field(default_factory=TokenUsage)
    total_duration: float = 0.0
    total_tool_calls: int = 0
    created_at: float = field(default_factory=time.time)


class MetricsCollector:
    """Collects and aggregates SDK metrics.
    
    Usage:
        collector = MetricsCollector()
        
        async with AsyncCodyClient(metrics=collector) as client:
            result = await client.run("task")
        
        # Get metrics
        metrics = collector.get_summary()
        print(f"Total tokens: {metrics.total_tokens.total_tokens}")
        print(f"Tool calls: {metrics.total_tool_calls}")
        print(f"Duration: {metrics.total_duration:.2f}s")
    """
    
    def __init__(self, max_runs: int = 1000):
        self._max_runs = max_runs
        self._runs: list[RunMetrics] = []
        self._tools: dict[str, ToolMetrics] = defaultdict(
            lambda: ToolMetrics(tool_name="unknown")
        )
        self._sessions: dict[str, SessionMetrics] = {}
        self._current_run: Optional[RunMetrics] = None
        self._start_time: float = 0.0
        self._enabled: bool = True
    
    def enable(self) -> None:
        """Enable metrics collection."""
        self._enabled = True
    
    def disable(self) -> None:
        """Disable metrics collection."""
        self._enabled = False
    
    def start_run(self, prompt: str, session_id: Optional[str] = None, thinking: bool = False) -> None:
        """Start tracking a run."""
        if not self._enabled:
            return
        
        self._current_run = RunMetrics(
            prompt_length=len(prompt),
            thinking_enabled=thinking,
            session_id=session_id,
        )
        self._start_time = time.time()
    
    def end_run(self, output: str, token_usage: TokenUsage) -> None:
        """End tracking a run."""
        if not self._enabled or not self._current_run:
            return
        
        self._current_run.output_length = len(output)
        self._current_run.duration = time.time() - self._start_time
        self._current_run.token_usage = token_usage
        
        self._runs.append(self._current_run)
        if len(self._runs) > self._max_runs:
            self._runs = self._runs[-self._max_runs:]

        # Update session metrics
        if self._current_run.session_id:
            session_id = self._current_run.session_id
            if session_id not in self._sessions:
                self._sessions[session_id] = SessionMetrics(session_id=session_id)
            
            session = self._sessions[session_id]
            session.run_count += 1
            session.total_tokens = session.total_tokens.add(token_usage)
            session.total_duration += self._current_run.duration
            session.total_tool_calls += self._current_run.tool_calls
        
        self._current_run = None
    
    def record_tool_call(
        self,
        tool_name: str,
        duration: float,
        success: bool = True,
        error: Optional[str] = None,
    ) -> None:
        """Record a tool call."""
        if not self._enabled:
            return
        
        if tool_name not in self._tools:
            self._tools[tool_name] = ToolMetrics(tool_name=tool_name)
        
        metrics = self._tools[tool_name]
        metrics.call_count += 1
        metrics.total_duration += duration
        
        if success:
            metrics.success_count += 1
        else:
            metrics.error_count += 1
        
        # Update current run
        if self._current_run:
            self._current_run.tool_calls += 1
    
    def record_token_usage(self, input_tokens: int, output_tokens: int) -> None:
        """Record token usage for current run."""
        if not self._enabled or not self._current_run:
            return
        
        self._current_run.token_usage = TokenUsage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=input_tokens + output_tokens,
        )
    
    def get_summary(self) -> dict:
        """Get aggregated metrics summary."""
        total_tokens = TokenUsage()
        total_duration = 0.0
        total_tool_calls = 0
        
        for run in self._runs:
            total_tokens = total_tokens.add(run.token_usage)
            total_duration += run.duration
            total_tool_calls += run.tool_calls
        
        return {
            "total_runs": len(self._runs),
            "total_tokens": total_tokens.total_tokens,
            "input_tokens": total_tokens.input_tokens,
            "output_tokens": total_tokens.output_tokens,
            "total_duration": total_duration,
            "total_tool_calls": total_tool_calls,
            "avg_run_duration": total_duration / len(self._runs) if self._runs else 0.0,
            "avg_tokens_per_run": total_tokens.total_tokens / len(self._runs) if self._runs else 0,
            "tool_metrics": {
                name: {
                    "call_count": m.call_count,
                    "success_rate": m.success_rate,
                    "avg_duration": m.avg_duration,
                }
                for name, m in self._tools.items()
            },
            "session_count": len(self._sessions),
        }
    
    def get_run_history(self) -> list[RunMetrics]:
        """Get history of all runs."""
        return list(self._runs)
    
    def get_tool_metrics(self, tool_name: Optional[str] = None) -> dict[str, ToolMetrics]:
        """Get tool metrics, optionally filtered by tool name."""
        if tool_name:
            return {tool_name: self._tools.get(tool_name, ToolMetrics(tool_name=tool_name))}
        return dict(self._tools)
    
    def get_session_metrics(self, session_id: str) -> Optional[SessionMetrics]:
        """Get metrics for a specific session."""
        return self._sessions.get(session_id)
    
    def reset(self) -> None:
        """Reset all metrics."""
        self._runs.clear()
        self._tools.clear()
        self._sessions.clear()
        self._current_run = None
    
    def export_json(self) -> dict:
        """Export metrics as JSON-serializable dict."""
        summary = self.get_summary()
        summary["runs"] = [
            {
                "prompt_length": r.prompt_length,
                "output_length": r.output_length,
                "duration": r.duration,
                "token_usage": {
                    "input": r.token_usage.input_tokens,
                    "output": r.token_usage.output_tokens,
                    "total": r.token_usage.total_tokens,
                },
                "tool_calls": r.tool_calls,
                "session_id": r.session_id,
            }
            for r in self._runs
        ]
        return summary
